# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module for building synaptic delay models"""

import os.path

import matplotlib.pyplot as plt
import numpy as np
import progressbar
from sklearn.linear_model import LinearRegression

from connectome_manipulator import log
from connectome_manipulator.model_building import model_types
from connectome_manipulator.access_functions import get_node_ids, get_edges_population, get_cv_data


def extract(
    circuit,
    bin_size_um,
    max_range_um=None,
    sel_src=None,
    sel_dest=None,
    sample_size=None,
    edges_popul_name=None,
    CV_dict=None,
    **_,
):
    """Extracts distance-dependent synaptic delays between samples of neurons.

    Args:
        circuit (bluepysnap.Circuit): Input circuit
        bin_size_um (float): Distance bin size in um
        max_range_um (float): Maximum distance range in um to consider
        sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
        sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
        sample_size (int): Size of random subsample of data to extract data from
        edges_popul_name (str): Name of SONATA egdes population to extract data from
        CV_dict (dict): Optional cross-validation dictionary, containing "n_folds" (int), "fold_idx" (int), "training_set" (bool) keys; will be automatically provided by the framework if "CV_folds" are specified

    Returns:
        dict: Dictionary containing the extracted data elements
    """
    # Select edge population
    edges = get_edges_population(circuit, edges_popul_name)

    # Select corresponding source/target nodes populations
    src_nodes = edges.source
    tgt_nodes = edges.target

    node_ids_src = get_node_ids(src_nodes, sel_src)
    node_ids_dest = get_node_ids(tgt_nodes, sel_dest)

    if sample_size is None or sample_size <= 0:
        sample_size = np.inf  # Select all nodes
    sample_size_src = min(sample_size, len(node_ids_src))
    sample_size_dest = min(sample_size, len(node_ids_dest))
    node_ids_src_sel = node_ids_src[
        np.random.permutation(
            [True] * sample_size_src + [False] * (len(node_ids_src) - sample_size_src)
        )
    ]
    node_ids_dest_sel = node_ids_dest[
        np.random.permutation(
            [True] * sample_size_dest + [False] * (len(node_ids_dest) - sample_size_dest)
        )
    ]

    # Cross-validation (optional)
    node_ids_src_sel, node_ids_dest_sel = get_cv_data(
        [node_ids_src_sel, node_ids_dest_sel], CV_dict
    )

    # Extract distance/delay values
    edges_table = edges.pathway_edges(
        source=node_ids_src_sel,
        target=node_ids_dest_sel,
        properties=[
            "@source_node",
            "delay",
            "afferent_center_x",
            "afferent_center_y",
            "afferent_center_z",
        ],
    )

    log.debug(
        f"Extracting delays from {edges_table.shape[0]} synapses (sel_src={sel_src}, sel_dest={sel_dest}, sample_size={sample_size} neurons, CV_dict={CV_dict})"
    )

    src_pos = src_nodes.positions(
        edges_table["@source_node"].to_numpy()
    ).to_numpy()  # Soma position of pre-synaptic neuron
    tgt_pos = edges_table[
        ["afferent_center_x", "afferent_center_y", "afferent_center_z"]
    ].to_numpy()  # Synapse position on post-synaptic dendrite
    src_tgt_dist = np.sqrt(np.sum((tgt_pos - src_pos) ** 2, 1))
    src_tgt_delay = edges_table["delay"].to_numpy()

    # Extract distance-dependent delays
    if max_range_um is None:
        max_range_um = np.max(src_tgt_dist)
    num_bins = np.ceil(max_range_um / bin_size_um).astype(int)
    dist_bins = np.arange(0, num_bins + 1) * bin_size_um
    dist_delays_mean = np.full(num_bins, np.nan)
    dist_delays_std = np.full(num_bins, np.nan)
    dist_count = np.zeros(num_bins).astype(int)

    log.debug("Extracting distance-dependent synaptic delays...")
    pbar = progressbar.ProgressBar()
    for idx in pbar(range(num_bins)):
        d_sel = np.logical_and(
            src_tgt_dist >= dist_bins[idx],
            (
                (src_tgt_dist < dist_bins[idx + 1])
                if idx < num_bins - 1
                else (src_tgt_dist <= dist_bins[idx + 1])
            ),
        )  # Including last edge
        dist_count[idx] = np.sum(d_sel)
        if dist_count[idx] > 0:
            dist_delays_mean[idx] = np.mean(src_tgt_delay[d_sel])
            dist_delays_std[idx] = np.std(src_tgt_delay[d_sel])

    return {
        "dist_bins": dist_bins,
        "dist_delays_mean": dist_delays_mean,
        "dist_delays_std": dist_delays_std,
        "dist_count": dist_count,
        "dist_delay_min": np.nanmin(src_tgt_delay),
    }


def build(dist_bins, dist_delays_mean, dist_delays_std, dist_delay_min, bin_size_um, **_):
    """Fits a linear distance-dependent synaptic delay model of type ``LinDelayModel`` to the data.

    Args:
        dist_bins (numpy.ndarray): Distance bin edges, as returned by :func:`extract`
        dist_delays_mean (numpy.ndarray): Delay mean for all bins, as returned by :func:`extract`
        dist_delays_std (numpy.ndarray): Delay std for all bins, as returned by :func:`extract`
        dist_delay_min (float): Overall delay minimum, as returned by :func:`extract`
        bin_size_um (float): Distance bin size in um

    Returns:
        connectome_manipulator.model_building.model_types.LinDelayModel: Fitted linear distance-dependent delay model
    """
    log.log_assert(np.all((np.diff(dist_bins) - bin_size_um) < 1e-12), "ERROR: Bin size mismatch!")
    bin_offset = 0.5 * bin_size_um

    # Mean delay model (linear)
    X = np.array(dist_bins[:-1][np.isfinite(dist_delays_mean)] + bin_offset, ndmin=2).T
    y = dist_delays_mean[np.isfinite(dist_delays_mean)]
    dist_delays_mean_fit = LinearRegression().fit(X, y)
    delay_mean_coeff_a = dist_delays_mean_fit.intercept_
    delay_mean_coeff_b = dist_delays_mean_fit.coef_[0]

    # Std delay model (const)
    delay_std = np.nanmean(dist_delays_std)

    # Min delay model (const)
    delay_min = dist_delay_min

    # Create model
    model = model_types.LinDelayModel(
        delay_mean_coeff_a=float(delay_mean_coeff_a),
        delay_mean_coeff_b=float(delay_mean_coeff_b),
        delay_std=float(delay_std),
        delay_min=float(delay_min),
    )
    log.debug("Model description:\n%s", model)

    return model


def plot(
    out_dir, dist_bins, dist_delays_mean, dist_delays_std, dist_count, model, **_
):  # pragma: no cover
    """Visualizes extracted data vs. actual model output.

    Args:
        out_dir (str): Path to output directory where the results figures will be stored
        dist_bins (numpy.ndarray): Distance bin edges, as returned by :func:`extract`
        dist_delays_mean (numpy.ndarray): Delay mean for all bins, as returned by :func:`extract`
        dist_delays_std (numpy.ndarray): Delay std for all bins, as returned by :func:`extract`
        dist_count (numpy.ndarray): Number of data elemets in each bin, as returned by :func:`extract`
        model (connectome_manipulator.model_building.model_types.LinDelayModel): Fitted linear distance-dependent delay model, as returned by :func:`build`
    """
    bin_width = np.diff(dist_bins[:2])[0]

    model_params = model.get_param_dict()
    mean_model_str = f'f(x) = {model_params["delay_mean_coeff_b"]:.3f} * x + {model_params["delay_mean_coeff_a"]:.3f}'
    std_model_str = f'f(x) = {model_params["delay_std"]:.3f}'
    min_model_str = f'f(x) = {model_params["delay_min"]:.3f}'

    # Draw figure
    model_kwargs = dict(zip(("src_type", "tgt_type"), model.default_types)) | {
        "distance": dist_bins
    }
    plt.figure(figsize=(8, 4), dpi=300)
    plt.bar(
        dist_bins[:-1] + 0.5 * bin_width,
        dist_delays_mean,
        width=0.95 * bin_width,
        facecolor="tab:blue",
        label=f"Data mean: N = {np.sum(dist_count)} synapses",
    )
    plt.bar(
        dist_bins[:-1] + 0.5 * bin_width,
        dist_delays_std,
        width=0.5 * bin_width,
        facecolor="tab:red",
        label=f"Data std: N = {np.sum(dist_count)} synapses",
    )
    plt.plot(
        dist_bins,
        model.get_mean(**model_kwargs),
        "--",
        color="tab:brown",
        label="Model mean: " + mean_model_str,
    )
    plt.plot(
        dist_bins,
        model.get_std(**model_kwargs),
        "--",
        color="tab:olive",
        label="Model std: " + std_model_str,
    )
    plt.plot(
        dist_bins,
        model.get_min(**model_kwargs),
        "--",
        color="tab:gray",
        label="Model min: " + min_model_str,
    )
    plt.xlim((dist_bins[0], dist_bins[-1]))
    plt.xlabel("Distance [um]")
    plt.ylabel("Delay [ms]")
    plt.title("Distance-dependent synaptic delays", fontweight="bold")
    plt.legend(loc="upper left", bbox_to_anchor=(1.1, 1.0))

    # Add second axis with bin counts
    count_color = "tab:orange"
    ax_count = plt.gca().twinx()
    ax_count.set_yscale("log")
    ax_count.step(dist_bins, np.concatenate((dist_count[:1], dist_count)), color=count_color)
    ax_count.set_ylabel("Count", color=count_color)
    ax_count.tick_params(axis="y", which="both", colors=count_color)
    ax_count.spines["right"].set_color(count_color)

    plt.tight_layout()

    out_fn = os.path.abspath(os.path.join(out_dir, "data_vs_model.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)

    # Visualize model output (generative model)
    dist_centers = [np.mean(dist_bins[:2]), np.mean(dist_bins), np.mean(dist_bins[-2:])]
    N = 1000  # Number of samples
    plt.figure(figsize=(8, 4), dpi=300)
    for didx, d in enumerate(dist_centers):
        plt.subplot(1, len(dist_centers), didx + 1)
        plt.hist(model.apply(distance=np.full(N, d)), bins=50)
        plt.ylim(plt.ylim())  # Freeze limit
        plt.title(f"{d:.0f} um")
        plt.xlabel("Delay [ms]")
        plt.ylabel("Count")
    plt.suptitle("Delay distributions")
    plt.tight_layout()

    out_fn = os.path.abspath(os.path.join(out_dir, "model_output.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import numpy as np
import os
import pandas as pd
import pytest

from bluepysnap import Circuit

from utils import TEST_DATA_DIR
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation
from connectome_manipulator.connectome_manipulation.converters import EdgeWriter
from connectome_manipulator import log


@pytest.fixture
def manipulation():
    m = Manipulation.get("syn_subsampling")
    return m


def test_apply(manipulation):
    log.setup_logging()  # To have data logging in a defined state

    c = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    edges = c.edges[c.edges.population_names[0]]
    nodes = [edges.source, edges.target]
    src_ids = nodes[0].ids()
    tgt_ids = nodes[1].ids()
    edges_table = edges.afferent_edges(tgt_ids, properties=edges.property_names)

    for pct in [10.1, 50.99999, 0.666, 99.9999]:
        writer = EdgeWriter(None, existing_edges=edges_table.copy())
        manipulation(nodes, writer).apply(tgt_ids, keep_pct=pct)
        res = writer.to_pandas()

        assert res.shape[0] == np.round(pct * edges_table.shape[0] / 100)
        assert np.all(
            [np.any(np.all(res.iloc[i] == edges_table, 1)) for i in range(res.shape[0])]
        )  # Check if all rows are contained



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""init"""

from connectome_manipulator.connectome_manipulation.manipulation.base import (
    Manipulation,
    MorphologyCachingManipulation,
)

from . import (
    conn_wiring,
    conn_rewiring,
    conn_extraction,
    conn_removal,
    null_manipulation,
    syn_subsampling,
    syn_removal,
    syn_prop_alteration,
)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Connectome (re)wiring module (general-purpose)."""

import libsonata
import neurom as nm
import numpy as np
import pandas as pd

from connectome_manipulator import log, profiler
from connectome_manipulator.access_functions import (
    get_node_ids,
    get_enumeration,
    get_node_positions,
)
from connectome_manipulator.connectome_manipulation.manipulation import (
    MorphologyCachingManipulation,
)
from connectome_manipulator.model_building import model_types, conn_prob

OPT_NCONN_MAX_ITER = 1000


class ConnectomeRewiring(MorphologyCachingManipulation):
    """Connectome manipulation class for (re)wiring a connectome:

    Rewires an existing connectome, or wires an empty connectome fom scratch, based on
    a given model of connection probability. Different aspects of connectivity can be
    preserved during rewiring.
    The manipulation can be applied through the :func:`apply` method.
    """

    # SONATA section type mapping (as in MorphIO): 1 = soma, 2 = axon, 3 = basal, 4 = apical
    SEC_SOMA = 1
    SEC_TYPE_MAP = {nm.AXON: 2, nm.BASAL_DENDRITE: 3, nm.APICAL_DENDRITE: 4}

    def __init__(self, nodes, writer, split_index=0, split_total=1):
        """Construct ConnectomeRewiring Manipulation and declare state vars..."""
        self.duplicate_sample_synapses_per_mtype_dict = tuple({})
        self.props_sel = []
        self.props_afferent = []
        self.syn_sel_idx_type = None
        super().__init__(nodes, writer, split_index, split_total)

    @profiler.profileit(name="conn_rewiring")
    def apply(
        self,
        split_ids,
        syn_class,
        prob_model_spec,
        delay_model_spec,
        sel_src=None,
        sel_dest=None,
        pos_map_file=None,
        keep_indegree=True,
        reuse_conns=True,
        gen_method=None,
        amount_pct=100.0,
        props_model_spec=None,
        nsynconn_model_spec=None,
        estimation_run=False,
        p_scale=1.0,
        opt_nconn=False,
        pathway_specs=None,
        keep_conns=False,
        rewire_mode=None,
        syn_pos_mode="reuse",
        syn_pos_model_spec=None,
        morph_ext="swc",
    ):
        """Applies a (re)wiring of connections between pairs of neurons based on a given connectivity model.

        Args:
            split_ids (list-like): List of neuron IDs that are part of the current data split; will be automatically provided by the manipulator framework
            syn_class (str): Selection of synapse class ("EXC" or "INH"), i.e., outgoing connections from either excitatory or inhibitory neuron types will be rewired at a time
            prob_model_spec (dict): Connection probability model specification; a file can be specified by ``{"file": "path/file.json"}``
            delay_model_spec (dict): Delay model specification; a file can be specified by ``{"file": "path/file.json"}``
            sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
            sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
            pos_map_file (str/list-like): Optional position mapping file pointing to a position mapping model (.json) or voxel data map (.nrrd); one or two files for source/target node populations may be provided
            keep_indegree (bool): If selected, the in-degree (number of incoming connections) of each rewired post-synaptic neuron is preserved
            reuse_conns (bool): If selected, existing (incoming) connections may be reused to form new connections during rewiring; specifically, synapses per connection, synapse positions, as well as synapse physiology are preserved, and only new pre-synaptic neurons are assigned to such connections
            gen_method (str): Method used for generating new synapses; can be "sample" (samples physiological property values independently from existing synapses) or "randomize" (draws physiological property values from model distributions; requires ``props_model_spec``); no ``gen_method`` required in case both ``keep_indegree`` and ``reuse_conns`` are selected
            amount_pct (float): Percentage of randomly sampled target (post-synaptic) neurons that will be wired
            props_model_spec (dict): Physiological properties model specification; must be provided for ``gen_method`` "randomize"; a file can be specified by ``{"file": "path/file.json"}``
            nsynconn_model_spec (dict): Model specifications for #synapses/connection; not required if #synapses/connection are part of ``props_model_spec``, but will override if still provided; a file can be specified by ``{"file": "path/file.json"}``
            estimation_run (bool): If selected, runs rewiring with early stopping, i.e., w/o generating an output connectome; an estimate of the average number of incoming connections for each post-synaptic neuron will be written to a data log file
            p_scale (float): Optional global probability scaling factor
            opt_nconn (bool): If selected, the number of ingoing connections for each post-neuron will be optimized to match its expected number of connections on average. This is done by repeating the random generation process up to ``OPT_NCONN_MAX_ITER=1000`` times and keeping the instance which has the exact or closest match
            pathway_specs (dict): Optional model specifications for efficiently setting model coefficients by pathway; will be automatically provided by the manipulator framework in case a .parquet file (containing a coefficient table for all pathways) is specified under "model_pathways" in the manipulation configuration file; only works with specific types of models
            keep_conns (bool): If selected, an existing connection is kept exactly as it is in case the same connection should be established during rewiring; otherwise, such a connection would be established by reusing another existing one (if ``reuse_conns`` selected) or by generating new synapses forming that connection
            rewire_mode (str): Optional selection of specific rewiring modes, such as "add_only" (only new connections can be added, all existing ones will be kept) or "delete_only" (only existing connections can be deleted, no new ones will be added); otherwise, there are no restrictions on rewiring, i.e., new connections may be added and existing ones deleted
            syn_pos_mode (str): Selection of synapse position mode for generating new synapses, such as "reuse" (reuses all existing synapse positions on the post-synaptic dendrites), "reuse_strict" (reuses only synapse positions on the post-synaptic dendrites that are incoming from the selected source neurons), "random" (randomly places new synapses on the actual dendritic morphologies; slower since access to morphologies is required), or "external" (synapse positions provided externally through ``syn_pos_model_spec``)
            syn_pos_model_spec (dict): External synapse position model specification of type ``PropsTableModel``; only required if ``syn_pos_mode`` "external" is selected; a file can be specified by ``{"file": "path/file.json"}``
            morph_ext (str): Morphology file extension, e.g., "swc", "asc", "h5"; only used if ``syn_pos_mode`` "random" is selected

        Note:
            Input/output edges (synapse) tables are accessed through the ``writer`` object:

            * Loading input edges: ``edges_table = self.writer.to_pandas()``
            * Writing output edges: ``self.writer.from_pandas(edges_table_manip)``

            The input edges table is assumed to be sorted by ""@target_node", and the
            output edges table will again be sorted by "@target_node".
        """
        # pylint: disable=arguments-differ
        edges_table = self.writer.to_pandas()
        log.log_assert(
            np.all(np.diff(edges_table["@target_node"]) >= 0),
            "Edges table must be ordered by @target_node!",
        )
        log.log_assert(
            syn_class in ["EXC", "INH"],
            f'Synapse class "{syn_class}" not supported (must be "EXC" or "INH")!',
        )
        log.log_assert(0.0 <= amount_pct <= 100.0, '"amount_pct" out of range!')
        log.log_assert(p_scale >= 0.0, '"p_scale" cannot be negative!')

        if estimation_run:
            log.log_assert(
                keep_indegree is False,
                'Connectivity estimation not supported with "keep_indegree" option!',
            )
            log.debug("*** Estimation run enabled ***")

        if opt_nconn:
            log.log_assert(
                keep_indegree is False,
                '#Connections optimization not supported with "keep_indegree" option!',
            )
            log.log_assert(
                estimation_run is False,
                '#Connections optimization not supported with "estimation_run" option!',
            )
            log.debug(
                f"Enabled optimization of #connections to match expected number on average (max. {OPT_NCONN_MAX_ITER} iterations)"
            )

        if rewire_mode is not None:
            log.log_assert(
                rewire_mode in ["add_only", "delete_only"],
                f'Rewire mode "{rewire_mode}" not supported (must be "add_only", "delete_only", or None for full rewiring)!',
            )
            log.log_assert(
                keep_indegree is False,
                f'"keep_indegree" not supported for rewire mode "{rewire_mode}"!',
            )

        log.log_assert(
            syn_pos_mode in ["reuse", "reuse_strict", "random", "external"],
            f'Synapse position mode "{syn_pos_mode}" not supported (must be "reuse", "reuse_strict", "random", or "external")!',
        )

        if keep_indegree and reuse_conns:
            log.log_assert(
                gen_method is None,
                'No generation method required for "keep_indegree" and "reuse_conns" options!',
            )
            log.log_assert(
                syn_pos_mode in ["reuse", "reuse_strict"],
                '"reuse[_strict]" synapse position mode required when using "keep_indegree" and "reuse_conns" options!',
            )
        else:
            log.log_assert(
                gen_method in ["sample", "randomize"],
                'Valid generation method required (must be "sample" or "randomize")!',
            )

        if gen_method == "sample" and self.split_total > 1:
            log.warning(
                f'Generation method "{gen_method}" samples only from synapses within same data split! Reduce number of splits to 1 to sample from all synapses!'
            )

        if "file" not in prob_model_spec:
            prob_model_spec["src_type_map"] = self.src_type_map
            prob_model_spec["tgt_type_map"] = self.tgt_type_map
            prob_model_spec["pathway_specs"] = pathway_specs

        # Load connection probability model
        p_model = model_types.AbstractModel.init_model(prob_model_spec)
        log.debug(f'Loaded conn. prob. model of type "{p_model.__class__.__name__}"')
        if p_scale != 1.0:
            log.debug(f"Using probability scaling factor p_scale={p_scale}")

        # Load delay model
        delay_model = model_types.AbstractModel.init_model(delay_model_spec)
        log.debug(f'Loaded delay model of type "{delay_model.__class__.__name__}"')

        # Load source/taget position mappings (optional; two types of mappings supported)
        pos_mappings = conn_prob.get_pos_mapping_fcts(pos_map_file)

        # Load connection/synapse properties model [required for "randomize" generation method]
        if gen_method == "randomize":
            log.log_assert(
                props_model_spec is not None,
                f'Properties model required for generation method "{gen_method}"!',
            )
            props_model = model_types.AbstractModel.init_model(props_model_spec)
            log.debug(f'Loaded properties model of type "{props_model.__class__.__name__}"')
            if nsynconn_model_spec is None:
                nsynconn_model = None
                log.log_assert(
                    props_model.has_nsynconn,
                    "#Syn/conn model required when using a properties model w/o nsynconn!",
                )
            else:
                nsynconn_model = model_types.AbstractModel.init_model(nsynconn_model_spec)
                log.debug(f'Loaded #syn/conn model of type "{nsynconn_model.__class__.__name__}"')
                if props_model.has_nsynconn:
                    log.warning(
                        "Separate #syn/conn model provided! #Syn/conn given by properties model will be ignored!",
                    )
        else:
            log.log_assert(
                props_model_spec is None,
                f'Properties model incompatible with generation method "{gen_method}"!',
            )
            props_model = None
            log.log_assert(
                nsynconn_model_spec is None,
                f'#Syn/conn model incompatible with generation method "{gen_method}"!',
            )
            nsynconn_model = None

        # Load synapse position model [required for "external" position mode]
        if syn_pos_mode == "external":
            log.log_assert(
                syn_pos_model_spec is not None,
                f'Synapse position model required for position mode "{syn_pos_mode}"!',
            )
            syn_pos_model = model_types.AbstractModel.init_model(syn_pos_model_spec)
            log.debug(f'Loaded synapse position model of type "{syn_pos_model.__class__.__name__}"')
        else:
            log.log_assert(
                syn_pos_model_spec is None,
                f'Synapse position model incompatible with position mode "{syn_pos_mode}"!',
            )
            syn_pos_model = None

        # Initialize statistics dict
        stats_dict = {}
        # Number of synapses and connections removed/rewired/added/kept
        stats_dict["num_syn_removed"] = []
        stats_dict["num_conn_removed"] = []
        stats_dict["num_syn_rewired"] = []
        stats_dict["num_conn_rewired"] = []
        stats_dict["num_syn_added"] = []
        stats_dict["num_conn_added"] = []
        stats_dict["num_syn_kept"] = []
        stats_dict["num_conn_kept"] = []
        # Total input synapse count
        stats_dict["input_syn_count"] = edges_table.shape[0]
        # Number of tgt neurons unable to rewire
        stats_dict["unable_to_rewire_nrn_count"] = 0
        # Number of input connections within src/tgt node selection
        stats_dict["input_conn_count_sel"] = []
        # Number of output connections within src/tgt node selection (based on prob. model; for specific seed)
        stats_dict["output_conn_count_sel"] = []
        # Average number of output connections within src/tgt node selection (based on prob. model)
        stats_dict["output_conn_count_sel_avg"] = []

        # Determine source/target nodes for rewiring
        src_node_ids = get_node_ids(self.nodes[0], sel_src)
        src_class = self.nodes[0].get(src_node_ids, properties="synapse_class")
        src_node_ids = src_class[
            src_class == syn_class
        ].index.to_numpy()  # Select only source nodes with given synapse class (EXC/INH)
        log.log_assert(len(src_node_ids) > 0, f"No {syn_class} source nodes found!")
        stats_dict["source_nrn_count_all"] = len(
            src_node_ids
        )  # All source neurons (corresponding to chosen sel_src and syn_class)
        syn_sel_idx_src = np.isin(edges_table["@source_node"], src_node_ids)
        log.log_assert(
            (
                np.all(edges_table.loc[syn_sel_idx_src, "syn_type_id"] >= 100)
                if syn_class == "EXC"
                else np.all(edges_table.loc[syn_sel_idx_src, "syn_type_id"] < 100)
            ),
            "Synapse class error!",
        )

        # Only select target nodes that are actually in current split of edges_table
        tgt_node_ids = get_node_ids(self.nodes[1], sel_dest, split_ids)
        num_tgt_total = len(tgt_node_ids)
        stats_dict["target_nrn_count_all"] = (
            num_tgt_total  # All target neurons in current split (corresponding to chosen sel_dest)
        )
        num_tgt = np.round(amount_pct * num_tgt_total / 100).astype(int)
        stats_dict["target_nrn_count_sel"] = (
            num_tgt  # Selected target neurons in current split (based on amount_pct)
        )
        tgt_sel = np.random.permutation([True] * num_tgt + [False] * (num_tgt_total - num_tgt))
        if num_tgt_total > 0:
            tgt_node_ids = tgt_node_ids[tgt_sel]  # Select subset of neurons (keeping order)
        if num_tgt == 0:  # Nothing to rewire
            log.debug("No target nodes selected, nothing to rewire")
            if estimation_run:
                log.data(
                    f"EstimationStats_{self.split_index + 1}_{self.split_total}",
                    input_syn_count=stats_dict["input_syn_count"],
                    source_nrn_count_all=stats_dict["source_nrn_count_all"],
                    target_nrn_count_all=stats_dict["target_nrn_count_all"],
                    target_nrn_count_sel=stats_dict["target_nrn_count_sel"],
                    unable_to_rewire_nrn_count=stats_dict["unable_to_rewire_nrn_count"],
                    input_conn_count_sel=stats_dict["input_conn_count_sel"],
                    output_conn_count_sel_avg=stats_dict["output_conn_count_sel_avg"],
                )
                self.writer.from_pandas(edges_table.iloc[[]].copy())
                return
            else:
                log.data(
                    f"RewiringIndices_{self.split_index + 1}_{self.split_total}",
                    i_split=self.split_index,
                    N_split=self.split_total,
                    split_ids=split_ids,
                    tgt_node_ids=tgt_node_ids,
                    tgt_sel=tgt_sel,
                )
                self.writer.from_pandas(edges_table)
                return

        # Get source/target node positions (optionally: two types of mappings)
        src_pos, tgt_pos = conn_prob.get_neuron_positions(
            self.nodes,
            [src_node_ids, tgt_node_ids],
            pos_acc=pos_mappings[0],
            vox_map=pos_mappings[1],
        )

        # Load target morphologies, if needed
        if syn_pos_mode == "random":
            tgt_morphs = self._get_tgt_morphs(morph_ext, libsonata.Selection(tgt_node_ids))
        else:  # "reuse", "reuse_strict", or "external"
            tgt_morphs = [None] * num_tgt

        log.info(
            f"Rewiring afferent {syn_class} connections to {num_tgt} ({amount_pct}%) of {len(tgt_sel)} target neurons in current split (total={num_tgt_total}, sel_src={sel_src}, sel_dest={sel_dest}, keep_indegree={keep_indegree}, gen_method={gen_method}, keep_conns={keep_conns}, reuse_conns={reuse_conns}, syn_pos_mode={syn_pos_mode}{', morph_ext=' + morph_ext if syn_pos_mode == 'random' else ''}, rewire_mode={rewire_mode})"
        )

        # Init/reset static variables (function attributes) related to generation methods which need only be initialized once [for better performance]
        self._reinit(edges_table, syn_class)
        # Index of input connections (before rewiring) [for data logging]
        inp_conns, inp_syn_conn_idx, inp_syn_per_conn = np.unique(
            edges_table[["@target_node", "@source_node"]],
            axis=0,
            return_inverse=True,
            return_counts=True,
        )
        inp_conns = np.fliplr(
            inp_conns
        )  # Restore ['@source_node', '@target_node'] order of elements
        stats_dict["input_conn_count"] = len(inp_syn_per_conn)
        stats_dict["input_syn_per_conn"] = list(inp_syn_per_conn)

        # Run connection rewiring
        syn_del_idx = np.full(
            edges_table.shape[0], False
        )  # Global synapse indices to keep track of all unused synapses to be deleted
        syn_rewire_idx = np.full(
            edges_table.shape[0], False
        )  # Global synapse indices to keep track of all rewired synapses [for data logging]
        new_edges_list = []  # New edges list to collect all generated synapses
        for tidx, (tgt, morph) in enumerate(zip(tgt_node_ids, tgt_morphs)):
            syn_sel_idx_tgt = edges_table["@target_node"] == tgt
            syn_sel_idx = np.logical_and(syn_sel_idx_tgt, syn_sel_idx_src)

            if syn_pos_mode == "reuse":
                syn_sel_idx_reuse = syn_sel_idx_tgt.copy()  # Reuse all synapses
            elif syn_pos_mode == "reuse_strict":
                syn_sel_idx_reuse = syn_sel_idx.copy()  # Reuse synapses restricted to sel_src
            else:
                syn_sel_idx_reuse = None

            if (keep_indegree and np.sum(syn_sel_idx) == 0) or (np.sum(syn_sel_idx_reuse) == 0):
                stats_dict["unable_to_rewire_nrn_count"] += 1  # (Neurons)
                # Nothing to rewire: either keeping indegree zero, or no target synapses exist
                # that could be rewired or positions reused from
                continue

            # Determine conn. prob. of all source nodes to be connected with target node
            p_src = (
                p_model.apply(
                    src_pos=src_pos,
                    tgt_pos=tgt_pos[tidx : tidx + 1, :],
                    src_type=get_enumeration(self.nodes[0], "mtype", src_node_ids),
                    tgt_type=get_enumeration(self.nodes[1], "mtype", [tgt]),
                    src_nid=src_node_ids,
                    tgt_nid=[tgt],
                ).flatten()
                * p_scale
            )
            p_src[np.isnan(p_src)] = 0.0  # Exclude invalid values
            p_src[src_node_ids == tgt] = (
                0.0  # Exclude autapses [ASSUMING node IDs are unique across src/tgt node populations!]
            )

            # Currently existing sources for given target node
            src, src_syn_idx = np.unique(
                edges_table.loc[syn_sel_idx, "@source_node"], return_inverse=True
            )
            num_src = len(src)
            stats_dict["input_conn_count_sel"] = stats_dict["input_conn_count_sel"] + [num_src]

            # Apply rewiring modes ("add_only", "delete_only", or full rewiring otherwise)
            conn_src = np.isin(src_node_ids, src)  # Existing source connectivity
            if rewire_mode == "add_only":  # Only new connections can be added, nothing deleted
                p_src = np.maximum(p_src, conn_src.astype(float))
            elif rewire_mode == "delete_only":  # Connections can only be deleted, nothing added
                p_src = np.minimum(p_src, conn_src.astype(float))

            # Sample new presynaptic neurons from list of source nodes according to conn. prob.
            if keep_indegree:  # Keep the same number of ingoing connections
                log.log_assert(
                    len(src_node_ids) >= num_src,
                    f"Not enough source neurons for target neuron {tgt} available for rewiring!",
                )
                log.log_assert(
                    np.sum(p_src) > 0.0,
                    "Keeping indegree not possible since connection probability zero!",
                )
                src_new = np.random.choice(
                    src_node_ids, size=num_src, replace=False, p=p_src / np.sum(p_src)
                )  # New source node IDs per connection
            else:  # Number of ingoing connections NOT necessarily kept the same
                stats_dict["output_conn_count_sel_avg"].append(np.round(np.sum(p_src)).astype(int))
                if estimation_run:
                    continue

                # Select source neurons (with or without optimizing numbers of connections)
                src_new_sel = self._select_sources(src_node_ids, p_src, opt_nconn)
                src_new = src_node_ids[src_new_sel]  # New source node IDs per connection
                stats_dict["output_conn_count_sel"].append(np.sum(src_new_sel))

            # Keep existing connections as they are (i.e., exclude them from any rewiring)
            if keep_conns:
                # Identify connections to keep
                keep_sel = np.logical_and(src_new_sel, conn_src)
                keep_ids = src_node_ids[keep_sel]  # Source node IDs to keep connections from

                # Remove from list of new connections
                src_new_sel[keep_sel] = False
                src_new = src_node_ids[src_new_sel]

                # Recompute source nodes selection used for rewiring
                keep_syn_idx = np.isin(edges_table["@source_node"], keep_ids)
                keep_syn_sel = np.logical_and(keep_syn_idx, syn_sel_idx)
                log.log_assert(
                    np.all(syn_sel_idx[keep_syn_sel]),
                    "ERROR: Inconsistent synapse indices to keep!",
                )
                syn_sel_idx[keep_syn_sel] = False

                src, src_syn_idx = np.unique(
                    edges_table.loc[syn_sel_idx, "@source_node"], return_inverse=True
                )
                num_src = len(src)

                # Synapse and connection statistics
                stats_dict["num_syn_kept"] = stats_dict["num_syn_kept"] + [np.sum(keep_syn_sel)]
                stats_dict["num_conn_kept"] = stats_dict["num_conn_kept"] + [len(keep_ids)]
            else:
                # Synapse and connection statistics
                stats_dict["num_syn_kept"] = stats_dict["num_syn_kept"] + [0]
                stats_dict["num_conn_kept"] = stats_dict["num_conn_kept"] + [0]

            # Randomize rewiring order of new source neurons
            src_new = np.random.permutation(src_new)
            num_new = len(src_new)

            # Re-use (up to) num_src existing connections (incl. #synapses/connection) for rewiring of (up to) num_new new connections (optional)
            if reuse_conns:
                num_src_to_reuse = num_src
                num_new_reused = num_new
            else:
                num_src_to_reuse = 0
                num_new_reused = 0

            if num_src > num_new_reused:  # Delete unused connections/synapses (randomly)
                src_syn_idx = self._shuffle_conns(src_syn_idx)
                syn_del_idx[syn_sel_idx] = (
                    src_syn_idx >= num_new_reused
                )  # Set global indices of connections to be deleted
                syn_sel_idx[syn_del_idx] = False  # Remove to-be-deleted indices from selection
                stats_dict["num_syn_removed"] = stats_dict["num_syn_removed"] + [
                    np.sum(src_syn_idx >= num_new_reused)
                ]
                stats_dict["num_conn_removed"] = stats_dict["num_conn_removed"] + [
                    num_src - num_new_reused
                ]
                src_syn_idx = src_syn_idx[src_syn_idx < num_new_reused]
            else:
                stats_dict["num_syn_removed"] = stats_dict["num_syn_removed"] + [0]
                stats_dict["num_conn_removed"] = stats_dict["num_conn_removed"] + [0]

            if num_src_to_reuse < num_new:  # Generate new synapses/connections, if needed
                num_gen_conn = num_new - num_src_to_reuse  # Number of new connections to generate
                src_gen = src_new[
                    -num_gen_conn:
                ]  # Split new sources into ones used for newly generated ...
                src_new = src_new[:num_src_to_reuse]  # ... and existing connections

                # Create new_edges and add them to global new edges table [ignoring duplicate indices]
                new_edges = self._generate_edges(
                    src_gen,
                    tidx,
                    tgt_node_ids,
                    syn_sel_idx_reuse,
                    edges_table,
                    gen_method,
                    props_model,
                    nsynconn_model,
                    delay_model,
                    morph,
                    syn_pos_model,
                )
                new_edges_list.append(new_edges)

                stats_dict["num_syn_added"] = stats_dict["num_syn_added"] + [new_edges.shape[0]]
                stats_dict["num_conn_added"] = stats_dict["num_conn_added"] + [len(src_gen)]
            else:
                stats_dict["num_syn_added"] = stats_dict["num_syn_added"] + [0]
                stats_dict["num_conn_added"] = stats_dict["num_conn_added"] + [0]

            # Assign new source nodes = rewiring of existing connections
            syn_rewire_idx = np.logical_or(syn_rewire_idx, syn_sel_idx)  # [for data logging]
            edges_table.loc[syn_sel_idx, "@source_node"] = src_new[
                src_syn_idx
            ]  # Source node IDs per connection expanded to synapses
            stats_dict["num_syn_rewired"] = stats_dict["num_syn_rewired"] + [len(src_syn_idx)]
            stats_dict["num_conn_rewired"] = stats_dict["num_conn_rewired"] + [len(src_new)]

            # Assign new distance-dependent delays (in-place), based on (generative) delay model
            self._assign_delays_from_model(
                delay_model, edges_table, src_new, src_syn_idx, syn_sel_idx
            )

        # Estimate resulting number of connections for computing a global probability scaling factor [returns empty edges table!!]
        if estimation_run:
            stat_sel = [
                "input_syn_count",
                "input_conn_count",
                "input_syn_per_conn",
                "source_nrn_count_all",
                "target_nrn_count_all",
                "target_nrn_count_sel",
                "unable_to_rewire_nrn_count",
                "input_conn_count_sel",
                "output_conn_count_sel_avg",
            ]
            stat_str = [
                (
                    f"      {k}: COUNT {len(v)}, MEAN {np.mean(v):.2f}, MIN {np.min(v)}, MAX {np.max(v)}, SUM {np.sum(v)}"
                    if isinstance(v, list) and len(v) > 0
                    else f"      {k}: {v}"
                )
                for k, v in stats_dict.items()
                if k in stat_sel
            ]
            log.debug("CONNECTIVITY ESTIMATION:\n%s", "\n".join(stat_str))
            log.data(
                f"EstimationStats_{self.split_index + 1}_{self.split_total}",
                **{k: v for k, v in stats_dict.items() if k in stat_sel},
            )
            self.writer.from_pandas(edges_table.iloc[[]].copy())
            return

        # Update statistics
        stats_dict["num_syn_unchanged"] = (
            stats_dict["input_syn_count"]
            - np.sum(stats_dict["num_syn_removed"])
            - np.sum(stats_dict["num_syn_rewired"])
        )

        # Delete unused synapses (if any)
        if np.any(syn_del_idx):
            edges_table = edges_table[~syn_del_idx].copy()
            log.debug(f"Deleted {np.sum(syn_del_idx)} unused synapses")

        # Add new synapses to table, re-sort, and assign new index
        if len(new_edges_list) > 0:
            all_new_edges = pd.concat(new_edges_list)
            syn_new_dupl_idx = np.array(
                all_new_edges.index
            )  # Index of duplicated synapses [for data logging]
            if edges_table.size == 0:
                max_idx = 0
            else:
                max_idx = np.max(edges_table.index)
            all_new_edges.index = range(
                max_idx + 1, max_idx + 1 + all_new_edges.shape[0]
            )  # Set index to new range, so as to keep track of new edges
            edges_table = pd.concat([edges_table, all_new_edges])
            edges_table.sort_values(
                "@target_node", kind="mergesort", inplace=True
            )  # Stable sorting, i.e., preserving order of input edges!!
            syn_new_idx = (
                edges_table.index > max_idx
            )  # Global synapse indices to keep track of all new synapses [for data logging]
            syn_new_dupl_idx = syn_new_dupl_idx[
                edges_table.index[syn_new_idx] - max_idx - 1
            ]  # Restore sorting, so that in same order as in merged & sorted edges table
            log.debug(f"Generated {all_new_edges.shape[0]} new synapses")
        else:  # No new synapses
            syn_new_dupl_idx = np.array([])
            syn_new_idx = np.full(edges_table.shape[0], False)

        # Reset index
        edges_table.reset_index(
            inplace=True, drop=True
        )  # Reset index [No index offset required when merging files in block-based processing]

        # [TESTING] #
        # Check if output indeed sorted
        log.log_assert(
            np.all(np.diff(edges_table["@target_node"]) >= 0),
            "ERROR: Output edges table not sorted by @target_node!",
        )
        # ######### #

        # Index of output connections (after rewiring) [for data logging]
        out_conns, out_syn_conn_idx, out_syn_per_conn = np.unique(
            edges_table[["@target_node", "@source_node"]],
            axis=0,
            return_inverse=True,
            return_counts=True,
        )
        out_conns = np.fliplr(
            out_conns
        )  # Restore ['@source_node', '@target_node'] order of elements
        stats_dict["output_syn_count"] = edges_table.shape[0]
        stats_dict["output_conn_count"] = len(out_syn_per_conn)
        stats_dict["output_syn_per_conn"] = list(out_syn_per_conn)

        # Log statistics
        stat_str = [
            (
                f"      {k}: COUNT {len(v)}, MEAN {np.mean(v):.2f}, MIN {np.min(v)}, MAX {np.max(v)}, SUM {np.sum(v)}"
                if isinstance(v, list) and len(v) > 0
                else f"      {k}: {v}"
            )
            for k, v in stats_dict.items()
        ]
        log.debug("STATISTICS:\n%s", "\n".join(stat_str))
        log.log_assert(
            stats_dict["num_syn_unchanged"]
            == stats_dict["output_syn_count"]
            - np.sum(stats_dict["num_syn_added"])
            - np.sum(stats_dict["num_syn_rewired"]),
            "ERROR: Unchanged synapse count mismtach!",
        )  # Consistency check
        log.data(f"RewiringStats_{self.split_index + 1}_{self.split_total}", **stats_dict)

        # Write index data log [book-keeping for validation purposes]
        inp_syn_unch_idx = np.zeros_like(
            syn_del_idx
        )  # Global synapse indices to keep track of all unchanged synapses [for data logging]
        inp_syn_unch_idx = np.logical_and(~syn_del_idx, ~syn_rewire_idx)
        out_syn_rew_idx = np.zeros_like(
            syn_new_idx
        )  # Global output synapse indices to keep track of all rewired synapses [for data logging]
        out_syn_rew_idx[~syn_new_idx] = syn_rewire_idx[
            ~syn_del_idx
        ]  # [ASSUME: Input edges table order preserved]
        out_syn_unch_idx = np.zeros_like(
            syn_new_idx
        )  # Global output synapse indices to keep track of all unchanged synapses [for data logging]
        out_syn_unch_idx[~syn_new_idx] = inp_syn_unch_idx[
            ~syn_del_idx
        ]  # [ASSUME: Input edges table order preserved]
        log.log_assert(
            np.sum(stats_dict["num_syn_rewired"]) == np.sum(syn_rewire_idx),
            "ERROR: Rewired (input) synapse count mismtach!",
        )
        log.log_assert(
            np.sum(stats_dict["num_syn_rewired"]) == np.sum(out_syn_rew_idx),
            "ERROR: Rewired (output) synapse count mismtach!",
        )
        log.log_assert(
            stats_dict["num_syn_unchanged"] == np.sum(inp_syn_unch_idx),
            "ERROR: Unchanged (input) synapse count mismtach!",
        )
        log.log_assert(
            stats_dict["num_syn_unchanged"] == np.sum(out_syn_unch_idx),
            "ERROR: Unchanged (output) synapse count mismtach!",
        )

        log.data(
            f"RewiringIndices_{self.split_index + 1}_{self.split_total}",
            inp_syn_del_idx=syn_del_idx,
            inp_syn_rew_idx=syn_rewire_idx,
            inp_syn_unch_idx=inp_syn_unch_idx,
            out_syn_new_idx=syn_new_idx,
            syn_new_dupl_idx=syn_new_dupl_idx,
            out_syn_rew_idx=out_syn_rew_idx,
            out_syn_unch_idx=out_syn_unch_idx,
            inp_conns=inp_conns,
            inp_syn_conn_idx=inp_syn_conn_idx,
            inp_syn_per_conn=inp_syn_per_conn,
            out_conns=out_conns,
            out_syn_conn_idx=out_syn_conn_idx,
            out_syn_per_conn=out_syn_per_conn,
            i_split=self.split_index,
            N_split=self.split_total,
            split_ids=split_ids,
            src_node_ids=src_node_ids,
            tgt_node_ids=tgt_node_ids,
            tgt_sel=tgt_sel,
        )
        # inp_syn_del_idx ... Binary index vector of deleted synapses w.r.t. input edges table (of current block)
        # inp_syn_rew_idx ... Binary index vector of rewired synapses w.r.t. input edges table (of current block)
        # inp_syn_unch_idx ... Binary index vector of unchanged synapses w.r.t. input edges table (of current block)
        # out_syn_new_idx ... Binary index vector of new synapses w.r.t. output edges table (of current block)
        # syn_new_dupl_idx ... Index vector of duplicated synapses (positions) w.r.t. input edges table (globally, i.e., across all blocks), corresponding to new synapses in out_syn_new_idx
        # out_syn_rew_idx ... Binary index vector of rewired synapses w.r.t. output edges table (of current block)
        # out_syn_unch_idx ... Binary index vector of unchanged synapses w.r.t. output edges table (of current block)
        # inp_conns ... Input connections (of current block)
        # inp_syn_conn_idx ... Index vector of input connections w.r.t. inp_conns (of current block)
        # inp_syn_per_conn: Number of synapses per connection w.r.t. inp_conns (of current block)
        # out_conns ... Input connections (of current block)
        # out_syn_conn_idx ... Index vector of input connections w.r.t. out_conns (of current block)
        # out_syn_per_conn ... Number of synapses per connection w.r.t. out_conns (of current block)
        # i_split ... Index of current block
        # N_split ... Total number of splits (blocks)
        # split_ids ... Neuron ids of current block
        # src_node_ids ... Selected source neuron ids
        # tgt_node_ids ... Selected target neuron ids within current block
        # tgt_sel ... Binary (random) target neuron selection index within current block, according to given amount_pct

        # [TESTING] #
        # Overflow/value check
        if edges_table.size > 0:
            log.log_assert(np.all(edges_table.abs().max() < 1e9), "Value overflow in edges table")
            if "n_rrp_vesicles" in edges_table.columns:
                log.log_assert(
                    np.all(edges_table["n_rrp_vesicles"] >= 1),
                    "Value error in edges table (n_rrp_vesicles)!",
                )
        # ######### #

        self.writer.from_pandas(edges_table)

    def _select_sources(self, src_node_ids, p_src, opt_nconn):
        """Select source neurons with or without optimizing numbers of connections."""
        if opt_nconn:
            # Optimizing #connections: Repeat random generation up to OPT_NCONN_MAX_ITER times
            #                          and keep the one with #connestions closest to average

            # Number of connections on average (=target count)
            num_conns_avg = np.round(np.sum(p_src)).astype(int)

            # Iterate OPT_NCONN_MAX_ITER times to find optimum
            new_conn_count = -np.inf
            for _ in range(OPT_NCONN_MAX_ITER):
                src_new_sel_tmp = np.random.rand(len(src_node_ids)) < p_src
                if np.abs(np.sum(src_new_sel_tmp) - num_conns_avg) < np.abs(
                    new_conn_count - num_conns_avg
                ):
                    # Keep closest value among all tries
                    src_new_sel = src_new_sel_tmp
                    new_conn_count = np.sum(src_new_sel)
                if new_conn_count == num_conns_avg:
                    break  # Optimum found
        else:  # Just draw once (w/o optimization)
            src_new_sel = np.random.rand(len(src_node_ids)) < p_src
        return src_new_sel

    def _shuffle_conns(self, syn_conn_idx):
        """Shuffles assignment of synapses to connections.

        e.g. [0, 0, 1, 1, 1, 2] -> [2, 2, 0, 0, 0, 1]
        """
        conn_map = np.random.permutation(np.max(syn_conn_idx) + 1)
        return conn_map[syn_conn_idx]

    def _generate_edges(
        self,
        src_gen,
        tidx,
        tgt_node_ids,
        syn_sel_idx_reuse,
        edges_table,
        gen_method,
        props_model,
        nsynconn_model,
        delay_model,
        morph,
        syn_pos_model,
    ):
        """Generates a new set of edges (=synapses), based on the chosen generation options.

        The generation method, use of morphologies, and delay model must be specified.
        """
        tgt = tgt_node_ids[tidx]

        # Create new synapses with pyhsiological parameters
        if gen_method == "sample":
            # Sample (non-morphology-related) property values independently from existing synapses
            new_edges, syn_conn_idx = self._create_synapses_by_sampling(
                src_gen, tidx, tgt_node_ids, edges_table
            )
        elif gen_method == "randomize":
            # Randomize (non-morphology-related) property values based on pathway-specific model distributions
            new_edges, syn_conn_idx = self._create_synapses_by_randomization(
                src_gen, tgt, props_model, nsynconn_model, edges_table
            )
        else:
            log.log_assert(False, f"Generation method {gen_method} unknown!")

        # Assign synapses to connections from src_gen to tgt
        # pylint: disable=E0601, E0606
        new_edges["@source_node"] = src_gen[syn_conn_idx]
        new_edges["@target_node"] = tgt

        # Fill-in synapse positions (in-place)
        if morph is None and syn_pos_model is None:  # i.e., syn_pos_mode "reuse" or "reuse_strict"
            # Duplicate synapse positions on target neuron
            self._reuse_synapse_positions(
                new_edges, edges_table, syn_sel_idx_reuse, syn_conn_idx, tgt
            )
        elif syn_pos_model is None and morph is not None:  # i.e., syn_pos_mode "random"
            # Randomly generate new synapse positions on target neuron
            self._generate_synapse_positions(morph, new_edges, syn_conn_idx)
        elif syn_pos_model is not None and morph is None:  # i.e., syn_pos_mode "external"
            # Load synapse positions externally from model
            self._load_synapse_positions(syn_pos_model, new_edges, src_gen, tgt)
        else:
            log.log_assert(False, "Synapse position mode error!")

        # Assign distance-dependent delays (in-place), based on (generative) delay model
        self._assign_delays_from_model(delay_model, new_edges, src_gen, syn_conn_idx)

        # Restore original data types
        new_edges = new_edges.astype(edges_table.dtypes)
        # new_edges = new_edges.astype(edges_table.dtypes[new_edges.columns])  # [ALTERNATIVE: In case of column mismatch!]

        return new_edges

    def _reuse_synapse_positions(
        self, new_edges, edges_table, syn_sel_idx_reuse, syn_conn_idx, tgt
    ):
        """Assigns (in-place) duplicate synapse positions on target neuron (w/o accessing dendritic morphologies).

        If possible, synapses will be selected such that no duplicated synapses belong to same connection.
        """
        if len(self.props_afferent) == 0:
            # No afferent properties to duplicate (i.e., point neurons)
            return

        conns, nsyns = np.unique(syn_conn_idx, return_counts=True)
        draw_from = np.where(syn_sel_idx_reuse)[0]
        sel_dupl = []
        unique_per_conn_warning = False
        for dupl_count in nsyns:
            if len(draw_from) >= dupl_count:
                sel_dupl.append(
                    np.random.choice(draw_from, dupl_count, replace=False)
                )  # Random sampling from existing synapses WITHOUT replacement, if possible
            else:
                sel_dupl.append(
                    np.random.choice(draw_from, dupl_count, replace=True)
                )  # Random sampling from existing synapses WITH replacement, otherwise
                unique_per_conn_warning = True
        sel_dupl = np.hstack(sel_dupl)

        if unique_per_conn_warning:
            log.warning(
                f"Duplicated synapse position belonging to same connection (target neuron {tgt})! Unique synapse positions per connection not possible!"
            )

        # [TESTING] #
        # Check if indeed no duplicates per connection
        if not unique_per_conn_warning:
            for cidx in conns:
                log.log_assert(
                    np.all(np.unique(sel_dupl[syn_conn_idx == cidx], return_counts=True)[1] == 1),
                    f"ERROR: Duplicated synapse positions within connection (target neuron {tgt})!",
                )
        # ######### #

        # Duplicate and assign afferent position properties
        new_edges[self.props_afferent] = edges_table.iloc[sel_dupl][self.props_afferent].to_numpy()

    def _generate_synapse_positions(self, morph, new_edges, syn_conn_idx):
        """Assign (in-place) new synapse positions on target neuron (with accessing dendritic morphologies).

        Synapses are randomly (uniformly) placed on soma/dendrite sections.
        Only afferent_... properties following the tuple representation (section_id, offset) will be set.
        """
        # Get available dendritic sections (plus soma) to place synapses on
        sec_ind = np.hstack(
            [
                [-1],  # [Soma]
                np.flatnonzero(
                    np.isin(morph.section_types, [nm.BASAL_DENDRITE, nm.APICAL_DENDRITE])
                ),
            ]
        )

        # Randomly choose section indices
        sec_sel = np.random.choice(sec_ind, len(syn_conn_idx))

        # Randomly choose fractional offset within each section
        off_sel = np.random.rand(len(syn_conn_idx))
        off_sel[sec_sel == -1] = 0.0  # Soma offsets must be zero

        # Synapse positions & (mapped) section types, computed from section & offset
        type_sel = np.full_like(sec_sel, self.SEC_SOMA)
        pos_sel = np.tile(morph.soma.center.astype(float), (len(sec_sel), 1))
        for idx in np.flatnonzero(sec_sel >= 0):
            type_sel[idx] = self.SEC_TYPE_MAP[morph.section(sec_sel[idx]).type]
            pos_sel[idx] = nm.morphmath.path_fraction_point(
                morph.section(sec_sel[idx]).points, off_sel[idx]
            )

        # Assign afferent position properties
        # IMPORTANT: Section IDs in NeuroM morphology don't include soma, so they need to be shifted by 1 (Soma ID is 0 in edges table)
        new_edges["afferent_section_id"] = sec_sel + 1
        new_edges["afferent_section_pos"] = off_sel
        new_edges["afferent_section_type"] = type_sel
        new_edges["afferent_center_x"] = pos_sel[:, 0]
        new_edges["afferent_center_y"] = pos_sel[:, 1]
        new_edges["afferent_center_z"] = pos_sel[:, 2]

    def _load_synapse_positions(self, syn_pos_model, new_edges, src_gen, tgt):
        """Assign (in-place) new synapse positions on target neuron (w/o accessing dendritic morphologies).

        Synapse positions are directly loaded from position table provided as PropsTableModel. An error is raised
        if not enough positions are available. No consistency checks against actual morphologies are done.
        Only afferent_... properties following the tuple representation (section_id, offset) will be set.
        """
        prop_names = [f"afferent_section_{_p}" for _p in ["id", "pos", "type"]] + [
            f"afferent_center_{_p}" for _p in ["x", "y", "z"]
        ]

        # Load synapse positions from model (order is arbitrary!)
        syn_pos = syn_pos_model.apply(
            src_nid=src_gen, tgt_nid=tgt, prop_names=["@source_node"] + prop_names
        ).set_index("@source_node")

        # Assign positions to source nodes
        for sid in src_gen:  # List of source node IDs
            conn_sel = new_edges["@source_node"] == sid
            n_sel = np.sum(conn_sel)
            new_edges.loc[conn_sel, prop_names] = syn_pos.loc[[sid]].to_numpy()[:n_sel, :]

    def _reinit(self, edges_table, syn_class):
        # Dict to keep computed values per target m-type (instead of re-computing them for each target neuron)
        self.duplicate_sample_synapses_per_mtype_dict = {}

        # Non-morphology-related property selection (to be sampled/randomized)
        self.props_sel = list(
            filter(
                lambda x: not np.any(
                    [
                        excl in x
                        for excl in [
                            "_node",
                            "_x",
                            "_y",
                            "_z",
                            "_section",
                            "_segment",
                            "_length",
                            "_morphology",
                            "delay",
                        ]
                    ]
                ),
                edges_table.columns,
            )
        )

        # Afferent morphology-related synapse properties (for duplicating synapses)
        self.props_afferent = list(filter(lambda nm: "afferent_" in nm, edges_table.columns))
        if len(self.props_afferent) == 0:
            log.warning('No "afferent_..." synapse properties - point neurons assumed!')

        # Synapse class selection (EXC or INH)
        if syn_class == "EXC":  # EXC: >=100
            self.syn_sel_idx_type = edges_table["syn_type_id"] >= 100
        elif syn_class == "INH":  # INH: 0-99
            self.syn_sel_idx_type = edges_table["syn_type_id"] < 100
        else:
            log.log_assert(False, f"Synapse class {syn_class} not supported!")

    def _create_synapses_by_sampling(self, src_gen, tidx, tgt_node_ids, edges_table):
        """Creates new synapses with pyhsiological parameter values by sampling.

        Works by sampling (non-morphology-related) property values, including numbers of
        synapses per connection, independently from existing synapses.
        All other properties will be initialized as zero (to be filled in later).
        """
        # Sample #synapses/connection from other existing synapses targetting neurons of the same mtype (or layer) as tgt (incl. tgt)
        tgt = tgt_node_ids[tidx]
        tgt_layers = self.nodes[1].get(tgt_node_ids, properties="layer").to_numpy()
        tgt_mtypes = self.nodes[1].get(tgt_node_ids, properties="mtype").to_numpy()
        tgt_mtype = tgt_mtypes[tidx]
        num_gen_conn = len(src_gen)
        if (
            tgt_mtype in self.duplicate_sample_synapses_per_mtype_dict
        ):  # Load from dict, if already exists [optimized for speed]
            syn_sel_idx_mtype = self.duplicate_sample_synapses_per_mtype_dict[tgt_mtype][
                "syn_sel_idx_mtype"
            ]
            num_syn_per_conn = self.duplicate_sample_synapses_per_mtype_dict[tgt_mtype][
                "num_syn_per_conn"
            ]
        else:  # Otherwise compute
            syn_sel_idx_mtype = np.logical_and(
                self.syn_sel_idx_type,
                np.isin(edges_table["@target_node"], tgt_node_ids[tgt_mtypes == tgt_mtype]),
            )
            if np.sum(syn_sel_idx_mtype) == 0:  # Ignore m-type, consider matching layer
                syn_sel_idx_mtype = np.logical_and(
                    self.syn_sel_idx_type,
                    np.isin(
                        edges_table["@target_node"], tgt_node_ids[tgt_layers == tgt_layers[tidx]]
                    ),
                )
            if np.sum(syn_sel_idx_mtype) == 0:  # Otherwise, ignore m-type & layer
                syn_sel_idx_mtype = self.syn_sel_idx_type
                log.warning(
                    f"No synapses with matching m-type or layer to sample connection property values for target neuron {tgt} from!"
                )
            log.log_assert(
                np.sum(syn_sel_idx_mtype) > 0,
                f"No synapses to sample connection property values for target neuron {tgt} from!",
            )
            _, num_syn_per_conn = np.unique(
                edges_table[syn_sel_idx_mtype][["@source_node", "@target_node"]],
                axis=0,
                return_counts=True,
            )
            self.duplicate_sample_synapses_per_mtype_dict[tgt_mtype] = {
                "syn_sel_idx_mtype": syn_sel_idx_mtype,
                "num_syn_per_conn": num_syn_per_conn,
            }
        num_syn_per_conn = num_syn_per_conn[
            np.random.choice(len(num_syn_per_conn), num_gen_conn)
        ]  # Sample #synapses/connection
        syn_conn_idx = np.concatenate(
            [[i] * n for i, n in enumerate(num_syn_per_conn)]
        )  # Create mapping from synapses to connections

        # Initialize new edges table with zeros (preserving data types)
        new_edges = pd.DataFrame(
            np.zeros((len(syn_conn_idx), len(edges_table.columns))), columns=edges_table.columns
        ).astype(edges_table.dtypes)

        # Sample (non-morphology-related) property values independently from other existing synapses targetting neurons of the same mtype as tgt (incl. tgt)
        # => Assume identical (non-morphology-related) property values for synapses belonging to same connection
        for p in self.props_sel:
            new_edges[p] = (
                edges_table.loc[syn_sel_idx_mtype, p]
                .sample(num_gen_conn, replace=True)
                .to_numpy()[syn_conn_idx]
            )

        return new_edges, syn_conn_idx

    def _create_synapses_by_randomization(
        self, src_gen, tgt, props_model, nsynconn_model, edges_table
    ):
        """Creates new synapses with pyhsiological parameter values by randomization.

        Works by randomly drawing (non-morphology-related) property values, including
        numbers of synapses per connection, from pathway-specific model distributions.
        All other properties will be initialized as zero (to be filled in later).
        """
        log.log_assert(
            np.all(np.isin(self.props_sel, props_model.get_prop_names())),
            f"Required properties missing in properties model (must include: {self.props_sel})!",
        )
        # Generate new synapse properties based on properties model
        src_mtypes = self.nodes[0].get(src_gen, properties="mtype").to_numpy()
        tgt_mtype = self.nodes[1].get(tgt, properties="mtype")
        if nsynconn_model is None:  # #Syn/conn part of props_model
            new_syn_props = [props_model.apply(src_type=s, tgt_type=tgt_mtype) for s in src_mtypes]
        else:  # Draw #syn/conn from nsynconn_model
            nsynconn = nsynconn_model.apply(src_nid=src_gen, tgt_nid=[tgt]).flatten()
            new_syn_props = [
                props_model.apply(src_type=s, tgt_type=tgt_mtype, n_syn=n)
                for s, n in zip(src_mtypes, nsynconn)
            ]
        num_syn_per_conn = [syn.shape[0] for syn in new_syn_props]
        syn_conn_idx = np.concatenate(
            [[i] * n for i, n in enumerate(num_syn_per_conn)]
        )  # Create mapping from synapses to connections

        # Initialize new edges table with zeros (preserving data types)
        new_edges = pd.DataFrame(
            np.zeros((len(syn_conn_idx), len(edges_table.columns))), columns=edges_table.columns
        ).astype(edges_table.dtypes)

        # Assign non-morphology-related property values
        new_edges[self.props_sel] = pd.concat(new_syn_props, ignore_index=True)[
            self.props_sel
        ].to_numpy()

        return new_edges, syn_conn_idx

    def _assign_delays_from_model(
        self, delay_model, edges_table, src_new, src_syn_idx, syn_sel_idx=None
    ):
        """Assign new distance-dependent delays, drawn from truncated normal distribution, to new synapses within edges_table (in-place)."""
        log.log_assert(delay_model is not None, "Delay model required!")

        if syn_sel_idx is None:
            syn_sel_idx = np.full(edges_table.shape[0], True)

        if len(src_new) == 0 or len(src_syn_idx) == 0 or np.sum(syn_sel_idx) == 0:
            # No synapses specified
            return

        # Determine distance from source neuron (soma) to synapse on target neuron
        # IMPORTANT: Distances for delays are computed in them original coordinate system w/o coordinate transformation!
        src_new_pos, _ = get_node_positions(self.nodes[0], src_new)
        # Synapse position on post-synaptic dendrite
        syn_pos = edges_table.loc[
            syn_sel_idx, ["afferent_center_x", "afferent_center_y", "afferent_center_z"]
        ].to_numpy()
        syn_dist = np.sqrt(np.sum((syn_pos - src_new_pos[src_syn_idx, :]) ** 2, 1))

        # Obtain delay values from (generative) model
        delay_new = delay_model.apply(distance=syn_dist)

        # Assign to edges_table (in-place)
        edges_table.loc[syn_sel_idx, "delay"] = delay_new



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

import os

import numpy as np
import pytest
import re
from bluepysnap import Circuit
from numpy.testing import assert_array_equal
from scipy.sparse import csc_matrix

from utils import TEST_DATA_DIR

import connectome_manipulator.connectome_comparison.adjacency as test_module


def test_compute():
    circuit = Circuit(os.path.join(TEST_DATA_DIR, "circuit_sonata.json"))
    nodes = [circuit.nodes["nodeA"]] * 2  # Src/tgt populations
    edges = circuit.edges["nodeA__nodeA__chemical"]
    edges_table = edges.afferent_edges(nodes[1].ids(), properties=edges.property_names)

    def get_adj(edges_table, src_ids, tgt_ids):
        """Extract adjacency matrix from edges table"""
        conns, cnts = np.unique(
            edges_table[["@source_node", "@target_node"]], axis=0, return_counts=True
        )
        adj_mat = np.zeros((len(src_ids), len(tgt_ids)), dtype=bool)
        cnt_mat = np.zeros((len(src_ids), len(tgt_ids)), dtype=int)
        for (_s, _t), _c in zip(conns, cnts):
            adj_mat[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = True
            cnt_mat[np.where(src_ids == _s)[0], np.where(tgt_ids == _t)[0]] = _c
        return adj_mat, cnt_mat

    def check_adj(res, ref_adj, ref_cnt, src_ids, tgt_ids):
        for _key in ["adj", "adj_cnt", "common"]:
            assert _key in res, f'ERROR: Results key "{_key}" missing!'
        for _key in ["src_gids", "tgt_gids"]:
            assert _key in res["common"], f'ERROR: Results key "{_key}" in "common" missing!'
        for _key in ["adj", "adj_cnt"]:
            assert "data" in res[_key], f'ERROR: Results key "data" in "{_key}" missing!'

        assert_array_equal(res["common"]["src_gids"], src_ids, "ERROR: Source IDs mismatch!")
        assert_array_equal(res["common"]["tgt_gids"], tgt_ids, "ERROR: Target IDs mismatch!")
        assert isinstance(res["adj"]["data"], csc_matrix), "ERROR: CSC matrix expected!"
        assert isinstance(res["adj_cnt"]["data"], csc_matrix), "ERROR: CSC matrix expected!"
        assert_array_equal(
            res["adj"]["data"].toarray(), ref_adj, "ERROR: Adjacency matrix mismatch!"
        )
        assert_array_equal(
            res["adj_cnt"]["data"].toarray(), ref_cnt, "ERROR: Count matrix mismatch!"
        )

    # Case 1: Invalid inputs
    ## (a) Invalid population name
    popul_name = "INVALID_POPULATION_NAME"
    with pytest.raises(
        AssertionError, match=re.escape(f'Population "{popul_name}" not found in edges file')
    ):
        res = test_module.compute(circuit, sel_src=None, sel_dest=None, edges_popul_name=popul_name)

    ## (b) Empty node sets
    popul_name = "nodeA__nodeA__chemical"
    for _src, _tgt in [(None, []), ([], None), ([], [])]:
        with pytest.raises(AssertionError, match=re.escape("Empty src/tgt node selection(s)")):
            res = test_module.compute(
                circuit, sel_src=_src, sel_dest=_tgt, edges_popul_name=popul_name
            )

    # Case 2: Full circuit
    res = test_module.compute(circuit, sel_src=None, sel_dest=None)
    ref_adj, ref_cnt = get_adj(edges_table, nodes[0].ids(), nodes[1].ids())
    check_adj(res, ref_adj, ref_cnt, nodes[0].ids(), nodes[1].ids())

    # Case 3: Partial circuit (layer by layer)
    for _src_lay in nodes[0].property_values("layer"):
        for _tgt_lay in nodes[1].property_values("layer"):
            sel_src = {"layer": _src_lay}
            sel_tgt = {"layer": _tgt_lay}
            res = test_module.compute(circuit, sel_src=sel_src, sel_dest=sel_tgt)
            ref_adj, ref_cnt = get_adj(edges_table, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt))
            check_adj(res, ref_adj, ref_cnt, nodes[0].ids(sel_src), nodes[1].ids(sel_tgt))



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module for alteration of synapse property values."""

import numpy as np

from connectome_manipulator import log, profiler
from connectome_manipulator.access_functions import get_node_ids
from connectome_manipulator.connectome_manipulation.manipulation import Manipulation


class SynapsePropertAlteration(Manipulation):
    """Connectome manipulation class for altering synapse property values:

    Modifies the values of a selected synapse property and a selected set of synapses,
    as given by constant, multiplicative, additive, or random values drawn from chosen
    distributions. The manipulation can be applied through the :func:`apply` method.
    """

    @profiler.profileit(name="syn_prop_alteration")
    def apply(
        self,
        split_ids,
        syn_prop,
        new_value,
        sel_src=None,
        sel_dest=None,
        syn_filter=None,
        amount_pct=100.0,
        **kwargs,
    ):
        """Applies a modification of synapse property values on a selected set of synapses.

        Args:
            split_ids (list-like): List of neuron IDs that are part of the current data split; will be automatically provided by the manipulator framework
            syn_prop (str): Selected synapse property values of which should be altered; can be selected from all available properties in the given SONATA input edges table (except for "@source_node" and "@target_node")
            new_value (dict): Dictionary specifying the type of alteration; must contain a "mode" key with one of the available alteration modes, together with additional settings depending on the mode, see Notes; optionally, can also contain a "range" key with lower/upper value bounds, an "rng" key to specify a random number generator, and a "kwargs" dict with additional arguments for the random number generator
            sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
            sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
            syn_filter (dict): Optional filter dictionary with property names (keys) with a corresponding list of values (e.g., ``{"layer": [2, 3, 4], "synapse_class": ["EXC"]}``); only synapses corresponding to that selection of values will be considered for alteration
            amount_pct (float): Percentage of randomly sampled synapses to be altered
            **kwargs: Additional keyword arguments - Not used

        Note:
            Available alteration modes and corresponding settings:

            * "setval": Absolute value, as given by "value"
            * "scale": Multiplicative scaling, as given by "factor"
            * "offset": Additive offset, as given by "value"
            * "shuffle": Shuffling of values among synapses; only supported when using a single data split
            * "randval": Random absolute value drawn from a given distribution (*)
            * "randscale": Multiplicative scaling by a random value drawn from a given distribution (*)
            * "randadd": Additive offset by a random value drawn from a given distribution (*)

            (*) A random distribution can be specified in ``new_value`` by choosing a random number generator as "rng" from ``numpy.random`` (e.g., "normal"), together with "kwargs" which will be passed when drawing values from that generator (e.g., "loc" and "scale" in case of "normal")

        Note:
            Input/output edges (synapse) tables are accessed through the ``writer`` object:

            * Loading input edges: ``edges_table = self.writer.to_pandas()``
            * Writing output edges: ``self.writer.from_pandas(edges_table_manip)``
        """
        # pylint: disable=arguments-differ
        if syn_filter is None:
            syn_filter = {}

        # Input checks
        edges_table = self.writer.to_pandas()
        available_properties = np.setdiff1d(
            edges_table.columns, ["@source_node", "@target_node"]
        ).tolist()  # Source/target nodes excluded
        available_modes = [
            "setval",
            "scale",
            "offset",
            "shuffle",
            "randval",
            "randscale",
            "randadd",
        ]  # Supported modes for generating new values
        log.log_assert(
            syn_prop in available_properties,
            f'syn_prop "{syn_prop}" not available or allowed for alteration! Must be one of: {available_properties}',
        )
        log.log_assert(
            np.all(np.isin(list(syn_filter.keys()), available_properties)),
            "One or more filter properties not available!",
        )
        log.log_assert(
            "mode" in new_value.keys(),
            f'Value "mode" must be specified! Available modes: {available_modes}',
        )
        log.log_assert(
            new_value["mode"] in available_modes,
            f'Value type "{new_value["mode"]}" unknown! Must be one of: {available_modes}',
        )
        log.log_assert(0.0 <= amount_pct <= 100.0, "amount_pct out of range!")

        # Select pathway synapses
        gids_src = get_node_ids(self.nodes[0], sel_src)
        gids_dest = get_node_ids(self.nodes[1], sel_dest)
        syn_sel_idx = np.logical_and(
            np.isin(edges_table["@source_node"], gids_src),
            np.isin(edges_table["@target_node"], gids_dest),
        )  # All potential synapses to be removed

        # Filter based on synapse properties (optional)
        if len(syn_filter) > 0:
            log.info(f"Applying synapse filter(s) on: {list(syn_filter.keys())}")
            for prop, val in syn_filter.items():
                syn_sel_idx = np.logical_and(syn_sel_idx, np.isin(edges_table[prop], val))

        # Apply alterations
        num_syn = np.sum(syn_sel_idx)
        num_alter = np.round(amount_pct * num_syn / 100).astype(int)

        log.info(
            f'Altering "{syn_prop}" in {num_alter} ({amount_pct}%) of {num_syn} selected synapses based on "{new_value["mode"]}" mode (sel_src={sel_src}, sel_dest={sel_dest})'
        )

        if num_alter < num_syn:
            sel_alter = np.random.permutation([True] * num_alter + [False] * (num_syn - num_alter))
            syn_sel_idx[syn_sel_idx] = sel_alter  # Set actual indices of synapses to be altered

        val_range = new_value.get("range", [-np.inf, np.inf])

        # Property data type to cast new values to, so that data type is not changed!!
        prop_dtype = edges_table.dtypes[syn_prop].type

        if new_value["mode"] == "setval":
            # Set to a fixed given value
            log.log_assert(
                new_value["value"] >= val_range[0] and new_value["value"] <= val_range[1],
                "Property value out of range!",
            )
            edges_table.loc[syn_sel_idx, syn_prop] = prop_dtype(new_value["value"])
        elif new_value["mode"] == "scale":
            # Scale by a given factor
            edges_table.loc[syn_sel_idx, syn_prop] = prop_dtype(
                np.clip(
                    edges_table.loc[syn_sel_idx, syn_prop] * new_value["factor"],
                    val_range[0],
                    val_range[1],
                )
            )
        elif new_value["mode"] == "offset":
            # Offset by a given value
            edges_table.loc[syn_sel_idx, syn_prop] = prop_dtype(
                np.clip(
                    edges_table.loc[syn_sel_idx, syn_prop] + new_value["value"],
                    val_range[0],
                    val_range[1],
                )
            )
        elif new_value["mode"] == "shuffle":
            # Shuffle across synapses
            log.log_assert(
                self.split_total == 1,
                f'"{new_value["mode"]}" mode not supported in block-based processing! Reduce number of splits to 1!',
            )
            edges_table.loc[syn_sel_idx, syn_prop] = edges_table.loc[syn_sel_idx, syn_prop].values[
                np.random.permutation(np.sum(syn_sel_idx))
            ]
        elif new_value["mode"] in ["randval", "randscale", "randadd"]:
            transform = {
                "randval": lambda _, y: y,
                "randscale": np.multiply,
                "randadd": np.add,
            }[new_value["mode"]]

            rng = getattr(np.random, new_value["rng"])
            random_values = rng(**new_value["kwargs"], size=np.sum(syn_sel_idx))
            new_values = np.clip(
                transform(edges_table.loc[syn_sel_idx, syn_prop], random_values),
                val_range[0],
                val_range[1],
            )

            # Set random values from given distribution
            rng = getattr(np.random, new_value["rng"])
            edges_table.loc[syn_sel_idx, syn_prop] = prop_dtype(new_values)
        else:
            log.log_assert(False, f'Value mode "{new_value["mode"]}" not implemented!')

        self.writer.from_pandas(edges_table)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Testing helpers"""

import pytest


@pytest.fixture
def nodes():
    class FakeNode:
        config = {"morphologies_dir": "/foo/bar"}
        _population = None

    return (FakeNode(), FakeNode())



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

from contextlib import redirect_stdout
import io
import os
import re
import logging

import pytest
from mock import patch

from utils import setup_tempdir
import connectome_manipulator.log as test_module


def test_log_assert():
    msg = "failing_assert_message"
    with patch("logging.log") as patched:
        with pytest.raises(AssertionError, match=msg):
            test_module.log_assert(False, msg)

        patched.assert_called()
        patched.reset_mock()

        test_module.log_assert(True, msg)
        patched.assert_not_called()


def test_logging_init():
    logname = "chronicle"
    with setup_tempdir(__name__) as tempdir:
        logdir = os.path.join(tempdir, "logs")
        test_module.setup_logging(logging.DEBUG)
        test_module.create_log_file(logdir, logname)

        assert os.path.isdir(logdir)

        # check directory has a file
        dir_listing = os.listdir(logdir)
        assert len(dir_listing) == 1

        # check that the file matches the naming policy
        assert re.match(rf"^{logname}.*\.log$", dir_listing[0])

        # check the log level is set correctly
        assert logging.getLogger().getEffectiveLevel() == logging.DEBUG



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module for building position mapping models based on pre-computed position tables"""

import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import progressbar
from scipy.spatial import distance_matrix

from connectome_manipulator import log
from connectome_manipulator.model_building import model_types
from connectome_manipulator.access_functions import get_node_ids


def extract(
    circuit,
    pos_file,
    coord_names,
    coord_scale=None,
    nodes_pop_name=None,
    nodes_spec=None,
    zero_based_indexing=False,
    gid_column=None,
    CV_dict=None,
    **_,
):
    """Loads pre-computed position mapping of a given nodes population.

    Args:
        circuit (bluepysnap.Circuit): Input circuit
        pos_file (str): Position table file name of format .feather, containing a pandas DataFrame
        coord_names (list-like): List of mapped coordinate names which must be present as columns in the DataFrame
        coord_scale (list-like): List of scaling factors for all mapped coordinates
        nodes_pop_name (str): Name of SONATA nodes population to extract data from
        nodes_spec (str/list-like/dict): Selection of neurons to be include in the mapping model
        zero_based_indexing (bool): If selected, zero-based indexing of neuron IDs in the position table file is assumed (as in SONATA); otherwise, one-based indexing is assumed (for backward compatibility)
        gid_column (str): Name of a column in the position table which contains the neuron IDs; if not provideed, the index column will be used
        CV_dict (dict): Cross-validation dictionary - Not supported

    Returns:
        dict: Dictionary containing the extracted data elements, i.e., neuron positions in original and mapped space
    """
    log.log_assert(CV_dict is None, "ERROR: Cross-validation not supported!")

    # Get neuron GIDs
    if nodes_pop_name is None:
        log.log_assert(
            len(circuit.nodes.population_names) == 1,
            f"ERROR: Nodes population could not be determined (found {circuit.nodes.population_names})!",
        )
        nodes_pop_name = circuit.nodes.population_names[0]
        log.debug(f'Loading nodes population "{nodes_pop_name}"')
    nodes = circuit.nodes[nodes_pop_name]

    nrn_ids = get_node_ids(nodes, nodes_spec)
    nrn_pos = nodes.positions(nrn_ids).to_numpy()  # Just for visualization
    nrn_lay = nodes.get(nrn_ids, properties="layer")  # Just for visualization
    if zero_based_indexing:
        idx_offset = 0  # Zero-based indexing in pos_file (SNAP/SONATA GIDs)
    else:
        idx_offset = 1  # One-based indexing in pos_file (BluePy GIDs)

    # Load position table
    log.log_assert(os.path.exists(pos_file), f'ERROR: Position file "{pos_file}" not found!')
    file_format = os.path.splitext(pos_file)[1]
    if file_format == ".feather":
        nrn_table = pd.read_feather(pos_file)
    else:
        log.log_assert(False, f'ERROR: "{file_format}" format not supported!')

    # pylint: disable=E0606
    log.debug(f'Loaded position table for {nrn_table.shape[0]} neurons from "{pos_file}"')

    # Assign mapped positions
    if gid_column is None:  # Use index column
        tab_gids = nrn_table.index.to_numpy()
    else:  # Use specified gid_column
        log.log_assert(
            gid_column in nrn_table.columns, f'ERROR: GID column "{gid_column}" not found!'
        )
        tab_gids = nrn_table[gid_column].to_numpy()

    log.log_assert(np.all(np.isin(nrn_ids + idx_offset, tab_gids)), "ERROR: Neuron IDs mismatch!")
    log.log_assert(
        np.all(np.isin(coord_names, nrn_table.columns)),
        "ERROR: Coordinate name(s) not found in position table!",
    )
    map_pos = nrn_table.loc[np.isin(tab_gids, nrn_ids + idx_offset), coord_names].to_numpy()

    # Apply scale
    if coord_scale is not None:
        log.log_assert(len(coord_names) == len(coord_scale), "ERROR: Coordinate scale mismatch!")
        map_pos = map_pos * coord_scale

    log.info(
        f'Loaded {", ".join(coord_names)} coordinates for {map_pos.shape[0]} neurons from position table (coord_scale={coord_scale})'
    )

    return {"nrn_ids": nrn_ids, "map_pos": map_pos, "nrn_pos": nrn_pos, "nrn_lay": nrn_lay}


def build(nrn_ids, coord_names, map_pos, model_coord_names=None, **_):
    """Builds a position mapping model from data.

    Args:
        nrn_ids (list-like): List of mapped neuron IDs, as returned by :func:`extract`
        coord_names (list-like): List of coordinate names of the mapped neuron positions table
        map_pos (numpy.ndarray): Mapped neuron positions table of size <#neurons x #coordinates>, as returned by :func:`extract`
        model_coord_names (list-like): Optional list of coordinate names that should be used in the resulting model

    Returns:
        connectome_manipulator.model_building.model_types.PosMapModel: Resulting position mapping model
    """
    if model_coord_names is None:
        model_coord_names = coord_names  # Same model coord names as input coord names
    else:
        log.log_assert(
            len(model_coord_names) == len(coord_names), "ERROR: Model coordinate names mismatch!"
        )

    map_pos_table = pd.DataFrame(map_pos, index=nrn_ids, columns=model_coord_names)
    log.log_assert(np.all(np.isfinite(map_pos_table)), "ERROR: Invalid mapped positions found!")

    # Create model
    model = model_types.PosMapModel(pos_table=map_pos_table)
    log.debug("Model description:\n%s", model)

    return model


def plot(out_dir, nrn_ids, nrn_pos, nrn_lay, model, **_):  # pragma: no cover
    """Visualizes neuron positions in original space vs. mapped space from model output.

    Args:
        out_dir (str): Path to output directory where the results figures will be stored
        nrn_ids (list-like): List of mapped neuron IDs, as returned by :func:`extract`
        nrn_pos (numpy.ndarray): Table of original neuron positions in 3D atlas space of size <#neurons x 3>, as returned by :func:`extract`
        nrn_lay (list-like): List of layer property values for all mapped neurons, as returned by :func:`extract`
        model (connectome_manipulator.model_building.model_types.PosMapModel): Resulting position mapping model, as returned by :func:`build`

    Note:
        Only mappings to 2D or 3D space are supported for visualization.
    """
    nrn_pos_model = model.apply(gids=nrn_ids)

    # Cell positions in 3D original vs. mapped space
    model_coord_names = model.get_coord_names()
    num_layers = len(np.unique(nrn_lay))
    lay_colors = plt.get_cmap("jet")(np.linspace(0, 1, num_layers))
    views_3d = [[90, 0], [0, 0]]
    pos_list = [nrn_pos, nrn_pos_model]
    lbl_list = ["Original space (data)", "Mapped space (model)"]
    coord_list = [["x [$\\mu$m]", "y [$\\mu$m]", "z [$\\mu$m]"], model_coord_names]
    fig = plt.figure(figsize=(10, 3 * len(views_3d)), dpi=300)
    plt.gcf().patch.set_facecolor("w")
    for vidx, v in enumerate(views_3d):
        for pidx, (pos, lbl, coo) in enumerate(zip(pos_list, lbl_list, coord_list)):
            if len(coo) == 2:
                if vidx == 0:  # Only plot once, since only single 2D view
                    ax = fig.add_subplot(
                        len(views_3d), len(pos_list), vidx * len(pos_list) + pidx + 1
                    )
                    for lidx in range(num_layers):
                        pos_sel = pos[nrn_lay == lidx + 1, :]
                        plt.plot(
                            pos_sel[:, 0],
                            pos_sel[:, 1],
                            ".",
                            color=lay_colors[lidx, :],
                            markersize=1.0,
                            alpha=0.5,
                            label=f"L{lidx + 1}",
                        )
                    ax.set_xlabel(coo[0])
                    ax.set_ylabel(coo[1])
            elif len(coo) == 3:
                ax = fig.add_subplot(
                    len(views_3d), len(pos_list), vidx * len(pos_list) + pidx + 1, projection="3d"
                )
                for lidx in range(num_layers):
                    pos_sel = pos[nrn_lay == lidx + 1, :]
                    plt.plot(
                        pos_sel[:, 0],
                        pos_sel[:, 1],
                        pos_sel[:, 2],
                        ".",
                        color=lay_colors[lidx, :],
                        markersize=1.0,
                        alpha=0.5,
                        label=f"L{lidx + 1}",
                    )
                ax.view_init(*v)
                ax.set_xlabel(coo[0])
                ax.set_ylabel(coo[1])
                ax.set_zlabel(coo[2])
            else:
                log.debug(f'Only 2D/3D plotting supported! Skipping "{lbl}"...')
                continue
            plt.legend(loc="center left", bbox_to_anchor=(1.0, 0.5), ncol=1)
            if vidx == 0:
                plt.title(lbl + f"\n[N={len(nrn_ids)}cells]")
    plt.tight_layout()

    out_fn = os.path.abspath(os.path.join(out_dir, "data_vs_model_positions.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)

    # Cell distances in atlas vs. mapped space
    max_plot = 10000
    if len(nrn_ids) > max_plot:
        log.debug("Using subsampling for distance plots!")
        nrn_sel = np.random.choice(len(nrn_ids), max_plot)
        nrn_ids = nrn_ids[nrn_sel]
        nrn_pos = nrn_pos[nrn_sel, :]
        nrn_pos_model = nrn_pos_model[nrn_sel, :]

    dist_mat_data = distance_matrix(nrn_pos, nrn_pos)
    dist_mat_model = distance_matrix(nrn_pos_model, nrn_pos_model)

    triu_idx = np.triu_indices(len(nrn_ids), 1)
    dist_val_data = dist_mat_data[triu_idx]
    dist_val_model = dist_mat_model[triu_idx]

    dist_max = max(dist_val_data, dist_val_model)
    plt.figure(figsize=(5, 5), dpi=300)
    plt.plot(dist_val_data, dist_val_model, "b.", alpha=0.1, markersize=1.0, markeredgecolor="none")
    plt.plot([0, dist_max], [0, dist_max], "k--")
    plt.xlim((0, dist_max))
    plt.ylim((0, dist_max))
    plt.grid(True)
    plt.xlabel("Distance in original space (data) [$\\mu$m]")
    plt.ylabel("Distance in mapped space (model)")
    plt.title(f"Cell distances in atlas vs. mapped space [N={len(nrn_ids)}cells]")
    plt.tight_layout()

    out_fn = os.path.abspath(os.path.join(out_dir, "data_vs_model_distances.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)

    # Nearest neighbors in atlas vs. mapped space
    NN_mat_data = np.argsort(dist_mat_data, axis=1)
    NN_mat_model = np.argsort(dist_mat_model, axis=1)

    num_NN_list = list(range(1, 30, 1))
    NN_match = np.full(len(num_NN_list), np.nan)

    log.debug("Computing nearest neighbors in atlas vs. mapped space...")
    pbar = progressbar.ProgressBar()
    for nidx in pbar(range(len(num_NN_list))):
        num_NN = num_NN_list[nidx]
        NN_match[nidx] = np.mean(
            [
                len(np.intersect1d(NN_mat_data[i, 1 : 1 + num_NN], NN_mat_model[i, 1 : 1 + num_NN]))
                / num_NN
                for i in range(len(nrn_ids))
            ]
        )

    plt.figure(figsize=(5, 4), dpi=300)
    plt.plot(num_NN_list, NN_match, ".-")
    plt.grid(True)
    plt.ylim((0, 1))
    plt.xlabel("#Nearest neighbors")
    plt.ylabel("Mean match")
    plt.title(f"Nearest neighbors in atlas vs. mapped space [N={len(nrn_ids)}cells]")

    out_fn = os.path.abspath(os.path.join(out_dir, "data_vs_model_neighbors.png"))
    log.info(f"Saving {out_fn}...")
    plt.savefig(out_fn)



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Manipulation base module.

Description: This module contains the Manipulation abstract base class of which all manipulation
classes must inherit and implement its methods.
"""

from abc import ABCMeta, abstractmethod
import functools
import inspect
import os.path

from bluepysnap.morph import MorphHelper
from bluepysnap.sonata_constants import Node
from morphio import Collection
import numpy as np

from connectome_manipulator import access_functions
from connectome_manipulator import log, profiler
from connectome_manipulator import utils
from connectome_manipulator.access_functions import get_enumeration_map


class MetaManipulation(ABCMeta):
    """Meta class to manage Manipulation algorithm classes.

    The only purpose of this Meta class derived from ABCMeta is to automatically register
    existing manipulation algorithms and give the programmer a helper function to load a class from
    an fct string in the config file.
    """

    __manipulations = {}

    def __init__(cls, name, bases, attrs) -> None:
        """Register the implementing class (if concrete) into a dictionary for later lookup"""
        if not inspect.isabstract(cls):
            modpath = inspect.getfile(cls)
            mod = os.path.splitext(os.path.basename(modpath))[0]
            cls.__manipulations[mod] = cls
        ABCMeta.__init__(cls, name, bases, attrs)

    @classmethod
    def get(mcs, name):
        """Returns a concrete Manipulation class given a string"""
        log.log_assert(name in mcs.__manipulations, f"Manipulation algorithm {name} does not exist")
        return mcs.__manipulations[name]


class Manipulation(metaclass=MetaManipulation):
    """Manipulation algorithm base class

    The abstract base class of which all manipulation classes must inherit and implement its methods.
    """

    def __init__(self, nodes, writer=None, split_index=0, split_total=1):
        """Initialize with the nodes and split_ids"""
        self.split_index = split_index
        self.split_total = split_total
        self.nodes = nodes
        self.writer = writer

        self.src_type_map = get_enumeration_map(self.nodes[0], "mtype")
        self.tgt_type_map = get_enumeration_map(self.nodes[1], "mtype")

    @abstractmethod
    def apply(self, split_ids, **kwargs):
        """An abstract method for the actual application of the algorithm

        This funciton is to be implemented by concrete Manipulation subclasses.
        """


class MorphologyCachingManipulation(Manipulation):
    """An abstract Manipulation with morphology caching

    This is a abstract Manipulation class that additionally provides a cache for morphologies,
    such that they can be reused on different invokations of apply without having to load them from
    the filesystem.
    """

    # pylint: disable=abstract-method

    def __init__(self, nodes, writer=None, split_index=0, split_total=1):
        """Initialize the MorphHelper object needed later."""
        super().__init__(nodes, writer, split_index, split_total)
        self.morpho_helper = MorphHelper(
            self.nodes[1].config.get("morphologies_dir"),
            self.nodes[1],
            self.nodes[1].config.get("alternate_morphologies"),
        )

    @profiler.profileit(name="morphology_reading")
    def _get_tgt_morphs(self, morph_ext, tgt_node_sel):
        """Access function (incl. transformation!), using specified format (swc/h5/...)"""
        result = access_functions.get_nodes(self.nodes[1], tgt_node_sel)
        return self._transform(
            self._get_cached_morphs(morph_ext, tuple(result[Node.MORPHOLOGY])),
            tgt_node_sel,
        )

    @functools.lru_cache(maxsize=10)
    def _get_cached_morphs(self, morph_ext, basenames):
        """Cached access function to get the untransformed morphologies."""
        morpho_dir = self.morpho_helper.get_morphology_dir(morph_ext)
        collection = Collection(morpho_dir, [f".{morph_ext}"])
        morphologies = [None] * len(basenames)
        for idx in collection.argsort(basenames):
            morphologies[idx] = collection.load(basenames[idx], mutable=True)
        return morphologies

    def _transform(self, morphs, node_sel):
        rotations = access_functions.orientations(self.nodes[1], node_sel)
        positions = access_functions.get_nodes(self.nodes[1], node_sel)
        positions = positions[[Node.X, Node.Y, Node.Z]].to_numpy()
        for m, r, p in zip(morphs, rotations, positions):
            T = np.eye(4)
            T[:3, :3] = r
            T[:3, 3] = p
            utils.transform(m, T)
            yield m.as_immutable()



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Customized logging"""

import logging
import os
import sys
from glob import glob
from time import strftime

import numpy as np

import connectome_manipulator

_LOG_FORMAT = "[%(levelname)s] %(message)s"
_LOG_FORMAT_WITH_DATE = "[%(levelname)s] (%(asctime)s) %(message)s"
_DATE_FORMAT = "%b.%d %H:%M:%S"

_start_time = None


def debug(msg, *args, **kwargs):  # pragma: no cover
    """Wrapper for debug logging"""
    return logging.debug(msg, *args, **kwargs)


def info(msg, *args, **kwargs):  # pragma: no cover
    """Wrapper for info logging"""
    return logging.info(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):  # pragma: no cover
    """Wrapper for warning logging"""
    return logging.warning(msg, *args, **kwargs)


def error(msg, *args, **kwargs):  # pragma: no cover
    """Wrapper for error logging"""
    return logging.error(msg, *args, **kwargs)


def exception(msg, *args, **kwargs):  # pragma: no cover
    """Wrapper for error logging"""
    return logging.exception(msg, *args, **kwargs)


def log_assert(cond, msg):
    """Assertion with logging"""
    if not cond:
        logging.log(logging.ERROR, msg)
    assert cond, msg


def data(filespec, **kwargs):
    """Data logging, i.e., writing data arrays given by kwargs to compressed .npz file

    WARNING: Existing files will be overwritten
    """
    file_handlers = [hdl for hdl in logging.root.handlers if hasattr(hdl, "baseFilename")]
    if len(file_handlers) == 0:
        warning("Data logging not possible!")
        return

    file_handler = file_handlers[0]
    base_name = os.path.splitext(file_handler.baseFilename)[0]
    if len(filespec) > 0:
        filespec = "." + filespec
    data_file = base_name + filespec + ".npz"
    if os.path.exists(data_file):
        # File may exist, e.g., in case of multiple rewiring functions.
        # Therefore, extend filename to prevent overwriting.
        p, df = os.path.split(data_file)
        fn, ext = os.path.splitext(df)
        file_list = list(glob(fn + ".*" + ext, root_dir=p))
        fidx = len(file_list) + 1
        data_file = os.path.join(p, fn + f".{fidx}" + ext)
        if os.path.exists(data_file):
            # Check again if for any reason this file still already exists
            warning(f'Data log file "{data_file}" already exists and will be overwritten!')
    np.savez_compressed(data_file, **kwargs)
    info(f'Data log ({", ".join(list(kwargs.keys()))}) written to "{data_file}"')


class LogLevel:
    """Class to select the log level."""

    ERROR_ONLY = 0
    DEFAULT = 1
    DEBUG = 2


def setup_logging(log_level=LogLevel.DEFAULT):
    """Features tabs and colors output to stdout

    Args:
      log_level (int): minimum log level for emitting messages
    """
    assert isinstance(log_level, int)
    log_level = min(log_level, 2)

    verbosity_levels = [
        logging.WARNING,  # pos 0: Minimum possible logging level
        logging.INFO,
        logging.DEBUG,
    ]

    # Stdout
    hdlr = logging.StreamHandler(sys.stdout)
    hdlr.setFormatter(logging.Formatter(_LOG_FORMAT, _DATE_FORMAT))
    hdlr.setLevel(verbosity_levels[log_level])
    logging.root.setLevel(verbosity_levels[log_level])

    del logging.root.handlers[:]
    logging.root.addHandler(hdlr)


def create_log_file(log_path, name):
    """Create the log file

    Args:
      log_path: The destination directory for log messages besides stdout
      name: Name of the module to log
    """
    global _start_time  # pylint: disable=global-statement
    os.makedirs(log_path, exist_ok=True)
    if not _start_time:
        _start_time = strftime("%Y-%m-%d_%Hh%M")
    logfile = os.path.join(log_path, "{}_{}.log".format(name, _start_time))
    fileh = logging.FileHandler(logfile, encoding="utf-8")
    fileh.setFormatter(logging.Formatter(_LOG_FORMAT_WITH_DATE, _DATE_FORMAT))
    logging.root.setLevel(logging.DEBUG)  # So that log files write everything
    remove = [h for h in logging.root.handlers if isinstance(h, logging.FileHandler)]
    for h in remove:
        logging.root.removeHandler(h)
    logging.root.addHandler(fileh)
    info(f'Log file "{logfile}" created!')
    info(f"Version: connectome_manipulator {connectome_manipulator.__version__}")
    return logfile



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Main module for structural connectome comparison:

- Loads two SONATA connectomes
- Extracts structural properties (compute_results), as specified by the structural comparator config dict (or re-loading them from disc, if computed earlier)
- Creates difference map between structural properties of the two connectomes
- Visualizes individual structural properties as well as their difference
"""

import importlib
import os.path
import pickle

import matplotlib.pyplot as plt
import numpy as np
from bluepysnap.circuit import Circuit

import connectome_manipulator


def compute_results(circuit, comp_dict):
    """Compute structural results used for circuit comparison."""
    comp_source = comp_dict["fct"]["source"]
    comp_kwargs = comp_dict["fct"]["kwargs"]

    comp_module = importlib.import_module(
        "connectome_manipulator.connectome_comparison." + comp_source
    )
    assert hasattr(
        comp_module, "compute"
    ), f'ERROR: Structural comparison module "{comp_source}" requires compute() function!'
    res_dict = comp_module.compute(circuit, **comp_kwargs)
    assert np.all(
        np.isin(comp_dict["res_sel"], list(res_dict.keys()))
    ), "ERROR: Specified results entry not found!"

    return res_dict


def results_diff(res_dict1, res_dict2):
    """Computes difference between two results data sets [recursively iterates through all sub-dicts to find 'data' entries; all other entries must be equal or will be removed]."""
    res_keys = np.intersect1d(list(res_dict1.keys()), list(res_dict2.keys()))
    for _k in np.setdiff1d(list(res_dict1.keys()), res_keys):
        del res_dict1[_k]
    for _k in np.setdiff1d(list(res_dict2.keys()), res_keys):
        del res_dict2[_k]
    assert np.all(
        [isinstance(res_dict1[k], type(res_dict2[k])) for k in res_keys]
    ), "ERROR: Results type mismatch!"

    diff_dict = {}
    for k in res_keys:
        if isinstance(res_dict1[k], dict):
            diff_dict[k] = results_diff(res_dict1[k], res_dict2[k])  # Recursive update
        else:
            if k == "data":
                assert (
                    res_dict1[k].shape == res_dict2[k].shape
                ), "ERROR: Results data shape mismatch!"
                # assert np.all(res_dict1[k] >= 0) and np.all(res_dict2[k] >= 0), 'ERROR: Negative results value(s) found!'
                diff_dict[k] = res_dict2[k].astype(float) - res_dict1[k].astype(
                    float
                )  # Difference matrix
            else:
                assert np.array_equal(res_dict1[k], res_dict2[k]), "ERROR: Results inconsistency!"
                diff_dict[k] = res_dict1[k]

    return diff_dict


def plot_results(res_dict, res_sel, plot_args, comp_dict):
    """Plot structural results."""
    comp_source = comp_dict["fct"]["source"]
    comp_kwargs = comp_dict["fct"]["kwargs"]

    comp_module = importlib.import_module(
        "connectome_manipulator.connectome_comparison." + comp_source
    )
    assert hasattr(
        comp_module, "plot"
    ), f'ERROR: Structural comparison module "{comp_source}" requires plot() function!'
    comp_module.plot(res_dict, res_sel, **plot_args, **comp_kwargs)


def main(structcomp_config, show_fig=False, force_recomp=False):  # pragma: no cover
    """Main entry point for structural connectome comparison."""
    print(f"VERSION INFO: connectome_manipulator {connectome_manipulator.__version__}")

    # Load circuits
    circuit_ids = sorted(structcomp_config["circuits"].keys())
    assert len(circuit_ids) == 2, "ERROR: Exactly two circuits required for comparison!"
    circuit_configs = [structcomp_config["circuits"][c]["circuit_config"] for c in circuit_ids]
    circuit_names = [structcomp_config["circuits"][c]["circuit_name"] for c in circuit_ids]

    circuits = [Circuit(cc) for cc in circuit_configs]
    print(f"INFO: {len(circuits)} circuits loaded:")
    for cc in circuit_configs:
        print("  " + cc)

    # Prepare saving
    out_dir = os.path.join(
        structcomp_config["out_dir"], f"{circuit_names[0]}_vs_{circuit_names[-1]}"
    )
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    # Computation & plotting
    if np.isscalar(force_recomp):
        force_recomp = [force_recomp] * len(circuit_ids)
    else:
        assert len(force_recomp) == len(
            circuit_ids
        ), f'ERROR: {len(circuit_ids)} "force_recomp" entries expected!'

    for plot_dict in structcomp_config["plot_types"]:
        print(f'INFO: Preparing "{plot_dict["name"]}" plot(s)...')

        # Load results if available, otherwise compute
        res_dicts = []
        for cidx in range(len(circuit_ids)):
            res_file = os.path.join(
                structcomp_config["working_dir"],
                "data",
                circuit_names[cidx] + "_" + plot_dict["name"] + ".pickle",
            )
            if os.path.exists(res_file) and not force_recomp[cidx]:
                # Load from file
                print(f"INFO: Loading results from {res_file}")
                with open(res_file, "rb") as f:
                    res_dict = pickle.load(f)
            else:
                # Compute & save to file
                res_dict = compute_results(circuits[cidx], plot_dict)
                print(f"INFO: Writing resutls to {res_file}")
                if not os.path.exists(os.path.split(res_file)[0]):
                    os.makedirs(os.path.split(res_file)[0])
                with open(res_file, "wb") as f:
                    pickle.dump(res_dict, f, protocol=4)
            res_dicts.append(res_dict)
        res_dicts.append(results_diff(res_dicts[0], res_dicts[-1]))

        def get_flattened_data(d):
            """Returns raw (flattened) data from sparse matrix or numpy array"""
            if hasattr(d, "flatten"):  # Numpy arrays
                return d.flatten()
            else:
                if hasattr(d, "data") and hasattr(d.data, "flatten"):  # Sparse matrix
                    return d.data.flatten()
                else:
                    assert False, "ERROR: Flattened data extraction error!"

        # Plot results
        for res_sel in plot_dict["res_sel"]:
            # Determine common range of values for plotting
            range_prctile = plot_dict.get("range_prctile", 100)
            all_data = np.concatenate(
                [
                    get_flattened_data(res_dicts[cidx][res_sel]["data"])
                    for cidx in range(len(circuit_ids))
                ]
            ).astype(float)
            all_data = all_data[np.isfinite(all_data)]
            plot_range = [
                (
                    -np.percentile(-all_data[all_data < 0], range_prctile)
                    if np.any(all_data < 0)
                    else 0.0
                ),
                (
                    np.percentile(all_data[all_data > 0], range_prctile)
                    if np.any(all_data > 0)
                    else 0.0
                ),
            ]  # Common plot range
            diff_data = get_flattened_data(res_dicts[-1][res_sel]["data"])
            plot_range_diff = max(
                (
                    np.percentile(diff_data[diff_data > 0], range_prctile)
                    if np.any(diff_data > 0)
                    else 0.0
                ),
                (
                    np.percentile(-diff_data[diff_data < 0], range_prctile)
                    if np.any(diff_data < 0)
                    else 0.0
                ),
            )  # Diff plot range
            if plot_range_diff == 0.0:
                plot_range_diff = 1.0  # Arbitrary range needed for symmetric plotting around 0.0 in case of zero difference
            plot_range_diff = [-plot_range_diff, plot_range_diff]  # Symmetric plot range

            # Create figure
            plt.figure(figsize=plot_dict.get("fig_size", None))
            num_subplots = len(res_dicts)
            for sidx in range(num_subplots):
                plt.subplot(1, num_subplots, sidx + 1)
                if sidx < len(circuit_ids):  # Separate results plots for each circuit
                    plot_args = {
                        "fig_title": circuit_names[sidx],
                        "vmin": plot_range[0],
                        "vmax": plot_range[-1],
                        "isdiff": False,
                    }
                else:  # Difference plot between circuits
                    plot_args = {
                        "fig_title": "Diff",
                        "vmin": plot_range_diff[0],
                        "vmax": plot_range_diff[-1],
                        "isdiff": True,
                    }
                plot_results(
                    res_dicts[sidx][res_sel], res_dicts[sidx]["common"], plot_args, plot_dict
                )

            plt.suptitle(res_dicts[-1][res_sel]["name"])
            plt.tight_layout()

            # Save figure
            fig_file = plot_dict.get("fig_file", {})
            out_fn = os.path.abspath(
                os.path.join(
                    out_dir,
                    f'struct_comp-{plot_dict["name"]}-{res_sel}.{fig_file.get("format", "pdf")}',
                )
            )
            print(f"INFO: Saving {out_fn}...")
            plt.gcf().savefig(out_fn, dpi=fig_file.get("dpi", None))

            if show_fig:
                plt.show()
            else:
                plt.close("all")



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Auxiliary classes to process many nodes split into chunks"""

from dataclasses import dataclass
from functools import reduce
from operator import and_
from typing import Union

import numpy as np
import pandas as pd

from . import log
from .access_functions import get_enumeration, get_enumeration_list


def _get_region_map(pop):
    """Get a mapping from region to node ids.

    Takes a bluepysnap population, will return a dataframe with hemisphere and region as
    index, node ids as column object.
    """
    hemis = get_enumeration(pop, "hemisphere")
    hemi_list = get_enumeration_list(pop, "hemisphere")
    regions = get_enumeration(pop, "region")
    region_list = get_enumeration_list(pop, "region")

    df = (
        pd.DataFrame(
            {
                "hemisphere": hemis,
                "region": regions,
                "id": np.arange(len(regions)),
            }
        )
        .groupby(["hemisphere", "region"])["id"]
        .apply(lambda x: x.values)
    )
    df.index = df.index.map(lambda x: (hemi_list[x[0]], region_list[x[1]]))
    return df


def _get_simple_node_splits(config, options, nodes):
    """Returns a simple, even split of target node ids"""
    if options.splits > 0:
        if "N_split_nodes" in config:
            log.debug(
                f"Overwriting N_split_nodes ({config['N_split_nodes']}) from configuration file with command line argument --split {options.splits}"
            )
        config["N_split_nodes"] = options.splits

    N_split = max(config.get("N_split_nodes", 1), 1)
    log.info(f"Setting up {N_split} processing batch jobs...")
    tgt_node_ids = nodes[1].ids()
    node_ids_split = np.split(
        tgt_node_ids, np.cumsum([np.ceil(len(tgt_node_ids) / N_split).astype(int)] * (N_split - 1))
    )
    return node_ids_split


def get_node_splits(config, options, nodes):
    """Split the target nodes into a series of lists of `BatchInfo` objects.

    If more than one manipulation is present in the passed configuration, or if no
    pathway-specific configuration is found, the resulting splitting is obtained by simply
    dividing the target nodes as evenly as possible.

    Otherwise, the pathway configuration is analyzed, and the target nodes are going to be
    grouped by region and hemisphere.  If a group is too large given the targeted workload
    defined by the options, the node ids within that group may be split. Similarly, if
    several groups together are smaller than the target workload, they may be grouped.

    The result is a list of lists containing `BatchInfo` objects, where each inner lists
    tries to have a workload as close as possible to the targeted value.

    Args:
        config: dictionary with settings for processing (the same as stored in JSON)
        options: command line options
        nodes: a tuple containing source and target node populations
    Returns:
        A list of `BatchInfo` groups splitting the passed in nodes
    """
    functions = config["manip"]["fcts"]
    if not (len(functions) == 1 and "model_pathways" in functions[0]):
        return [
            [BatchInfo(None, None, ids)] for ids in _get_simple_node_splits(config, options, nodes)
        ]

    src_ids = _get_region_map(nodes[0])
    dst_ids = _get_region_map(nodes[1])

    filename = options.config_path.parent / functions[0]["model_pathways"]
    pathways = pd.read_parquet(filename)

    payloads = (
        pathways.index.unique()
        .to_frame(index=False)
        .join(src_ids.map(len), on=[f"src_{c}" for c in src_ids.index.names])
        .groupby(["dst_hemisphere", "dst_region"])
        .sum(numeric_only=True)
    )

    # Consider two queues: jobs that split a region go into full_batches, whereas other
    # regions are too small to be split and can be grouped further and end up in
    # partial_batches.
    full_batches = []
    partial_batches = []
    for (dst_hemisphere, dst_region), (payload,) in payloads.iterrows():
        dst_count = max(1, int(options.target_payload / payload))
        ids = dst_ids.loc[dst_hemisphere, dst_region]
        sel = {"hemisphere": dst_hemisphere, "region": dst_region}
        if len(ids) <= dst_count:
            # Don't split, merge later
            partial_batches.append(BatchInfo(len(ids) * payload, sel, ids))
        else:
            splits = (len(ids) + dst_count - 1) // dst_count
            for split in np.array_split(ids, splits):
                # Single full batch, to be processed by itself
                full_batches.append([BatchInfo(len(split) * payload, sel, split)])
    # Groups partial batches
    full_batches.extend(BatchInfo.group_batches(partial_batches, options.target_payload))

    return full_batches


@dataclass
class BatchInfo:
    """An abstraction for a chunk of work"""

    payload: Union[int, None]
    """How much work this batch does, counting the number of src-node connections"""

    selection: Union[dict, None]
    """A SONATA nodest-style selection akin to property: value"""

    node_ids: list
    """Either a plain list of ids or a list of SONATA selection style ranges"""

    def __repr__(self):
        """Condensed representation of the class"""
        if len(self.node_ids) > 5:
            node_str = f"[{self.node_ids[0]}, ..., {self.node_ids[-1]}]"
        else:
            node_str = str(self.node_ids)
        return f"BatchInfo(payload={self.payload}, selection={self.selection}, node_ids={node_str})"

    @staticmethod
    def group_batches(all_batches: ["BatchInfo"], target_payload: int) -> [["BatchInfo"]]:
        """Partition batches into groups given a desired target_payload"""
        jobs = []
        while all_batches:
            batches = [all_batches.pop(0)]
            payload = batches[-1].payload
            while all_batches and target_payload >= payload + all_batches[0].payload:
                batches.append(all_batches.pop(0))
                payload += batches[-1].payload
            jobs.append(batches)
        return jobs

    def process_pathways(self, pathways=None):
        """Iterates over relevant pathway configurations for this batch

        Takes a Pandas Dataframe and yields relevant node ids, source and target selection
        dictionaries, as well as the pathway group configuration.
        """
        if pathways is None:
            yield self.node_ids, None, self.selection, None
            return

        if self.selection:
            index = pathways.index.to_frame()
            sel = []
            for k, v in self.selection.items():
                sel.append(index[f"dst_{k}"] == v)
            pathways = pathways[reduce(and_, sel)]

        grouped = pathways.groupby(pathways.index.names, observed=True)
        for idx, group in grouped:
            sel_src = {}
            sel_dst = {}
            for k, v in zip(pathways.index.names, idx):
                if k.startswith("src_"):
                    sel_src[k[4:]] = v
                elif k.startswith("dst_"):
                    sel_dst[k[4:]] = v
                else:
                    raise ValueError(f"Incompatible pathway index column: {k}")
            group = group.set_index(["src_type", "dst_type"])
            yield self.node_ids, sel_src, sel_dst, group



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module for comparing connectomes based on synapse properties:

Structural comparison of two connectomes in terms of synapse properties per pathway,
as specified by the config. For each connectome, the underlying properties maps are
computed by the :func:`compute` function and will be saved to a data file first.
The individual synapse properties maps, together with a difference map between the
two connectomes, are then plotted by means of the :func:`plot` function.
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import progressbar
from connectome_manipulator.access_functions import get_edges_population, get_node_ids


def compute(
    circuit,
    fct="np.mean",
    group_by=None,
    sel_src=None,
    sel_dest=None,
    per_conn=False,
    skip_empty_groups=False,
    edges_popul_name=None,
    **_,
):
    """Computes a matrix of synapse property values between groups of neurons of a given circuit's connectome.

    Args:
        circuit (bluepysnap.Circuit): Input circuit
        fct (str): Function to apply, e.g., "np.mean", "np.std"
        group_by (str): Neuron property name based on which to group connections, e.g., "synapse_class", "layer", or "mtype"; if omitted, the overall average is computed
        sel_src (str/list-like/dict): Source (pre-synaptic) neuron selection
        sel_dest (str/list-like/dict): Target (post-synaptic) neuron selection
        per_conn (bool): If selected, ``fct`` is applied to the average property value per connection (i.e., average value of all synapses belonging to a connection); otherwise, ``fct`` is applied to the synapses of all connections altogether
        skip_empty_groups (bool): If selected, only group property values that exist within the given source/target selection are kept; otherwise, all group property values, even if not present in the given source/target selection, will be included
        edges_popul_name (str): Name of SONATA egdes population to extract data from

    Returns:
        dict: Dictionary containing the computed data elements; see Notes

    Note:
        The returned dictionary contains the data elements that can be selected for plotting through the structural comparison configuration file, together with a common dictionary containing additional information. Each data element is a dictionary with "data" (numpy.ndarray of size <source-group-size x target-group-size>), "name" (str), and "unit" (str) items. Names of these data elements correspond to the synapse properties that are present in the given SONATA edges population. Usual properties may include for example:

        * "conductance": Peak conductance
        * "decay_time": Decay time constant
        * "depression_time": Time constant for recovery from depression
        * "facilitation_time": Time constant for recovery from facilitation
        * "u_syn": Utilization of synaptic efficacy
        * ...
    """
    # Select edge population
    edges = get_edges_population(circuit, edges_popul_name)

    # Select corresponding source/target nodes populations
    src_nodes = edges.source
    tgt_nodes = edges.target

    if group_by is None:
        src_group_sel = [sel_src]
        tgt_group_sel = [sel_dest]
    else:
        if (
            skip_empty_groups
        ):  # Take only group property values that exist within given src/tgt selection
            src_group_values = np.unique(
                src_nodes.get(get_node_ids(src_nodes, sel_src), properties=group_by)
            )
            tgt_group_values = np.unique(
                tgt_nodes.get(get_node_ids(tgt_nodes, sel_dest), properties=group_by)
            )
        else:  # Keep all group property values, even if not present in given src/tgt selection, to get the full matrix
            src_group_values = sorted(src_nodes.property_values(group_by))
            tgt_group_values = sorted(tgt_nodes.property_values(group_by))

        if sel_src is None:
            sel_src = {}
        else:
            assert isinstance(
                sel_src, dict
            ), "ERROR: Source node selection must be a dict or empty!"  # Otherwise, it cannot be merged with group selection
        if sel_dest is None:
            sel_dest = {}
        else:
            assert isinstance(
                sel_dest, dict
            ), "ERROR: Target node selection must be a dict or empty!"  # Otherwise, it cannot be merged with pathway selection

        src_group_sel = [
            {**sel_src, group_by: src_group_values[idx]} for idx in range(len(src_group_values))
        ]  # group_by will overwrite selection in case group property also exists in selection!
        tgt_group_sel = [
            {**sel_dest, group_by: tgt_group_values[idx]} for idx in range(len(tgt_group_values))
        ]  # group_by will overwrite selection in case group property also exists in selection!

    print(
        f"INFO: Extracting synapse properties (group_by={group_by}, sel_src={sel_src}, sel_dest={sel_dest}, N={len(src_group_values)}x{len(tgt_group_values)} groups, per_conn={per_conn})",
        flush=True,
    )

    edge_props = sorted(edges.property_names)
    print(f"INFO: Available synapse properties: \n{edge_props}", flush=True)

    prop_fct = eval(fct)
    prop_tables = np.full((len(src_group_sel), len(tgt_group_sel), len(edge_props)), np.nan)
    pbar = progressbar.ProgressBar()
    for idx_pre in pbar(range(len(src_group_sel))):
        sel_pre = src_group_sel[idx_pre]
        for idx_post, _ in enumerate(tgt_group_sel):
            sel_post = tgt_group_sel[idx_post]
            pre_ids = get_node_ids(src_nodes, sel_pre)
            post_ids = get_node_ids(tgt_nodes, sel_post)
            e_sel = edges.pathway_edges(pre_ids, post_ids, edge_props)
            if e_sel.size > 0:
                if per_conn:  # Apply prop_fct to average value per connection
                    conn, conn_idx = np.unique(
                        e_sel[["@source_node", "@target_node"]], axis=0, return_inverse=True
                    )
                    c_sel = pd.DataFrame(index=range(conn.shape[0]), columns=edge_props)
                    for cidx in range(conn.shape[0]):
                        c_sel.loc[cidx, :] = np.mean(e_sel[conn_idx == cidx], axis=0)
                    prop_tables[idx_pre, idx_post, :] = prop_fct(c_sel.to_numpy(), axis=0)
                else:
                    prop_tables[idx_pre, idx_post, :] = prop_fct(e_sel.to_numpy(), axis=0)

    fname = prop_fct.__name__[0].upper() + prop_fct.__name__[1:]
    cname = " (per conn)" if per_conn else ""
    res_dict = {
        edge_props[idx]: {
            "data": prop_tables[:, :, idx],
            "name": f'"{edge_props[idx]}" property',
            "unit": f"{fname} {edge_props[idx]}{cname}",
        }
        for idx in range(len(edge_props))
    }
    res_dict["common"] = {
        "src_group_values": src_group_values,
        "tgt_group_values": tgt_group_values,
    }

    return res_dict


def plot(
    res_dict, common_dict, fig_title=None, vmin=None, vmax=None, isdiff=False, group_by=None, **_
):  # pragma:no cover
    """Plots a properties matrix or a difference matrix.

    Args:
        res_dict (dict): Results dictionary, containing selected data for plotting; must contain a "data" item with a properties matrix of type numpy.ndarray of size  <#source-group-values x #target-group-values>, as well as "name" and "unit" items containing strings.
        common_dict (dict): Common dictionary, containing additional information; must contain "src_group_values" and "tgt_group_values" items containing lists of source/target values of the grouped property, matching the size of the properties matrix in ``res_dict``
        fig_title (str): Optional figure title
        vmin (float): Minimum plot range
        vmax (float): Maximum plot range
        isdiff (bool): Flag indicating that ``res_dict`` contains a difference matrix; in this case, a symmetric plot range is required and a divergent colormap will be used
        group_by (str): Neuron property name based on which to group connections, e.g., "synapse_class", "layer", or "mtype"; if omitted, the overall average is computed
    """
    if isdiff:  # Difference plot
        assert -1 * vmin == vmax, "ERROR: Symmetric plot range required!"
        cmap = "PiYG"  # Symmetric (diverging) colormap
    else:  # Regular plot
        cmap = "hot_r"  # Regular colormap

    plt.imshow(res_dict["data"], interpolation="nearest", cmap=cmap, vmin=vmin, vmax=vmax)

    if fig_title is None:
        plt.title(res_dict["name"])
    else:
        plt.title(fig_title)

    if group_by:
        plt.xlabel(f"Postsynaptic {group_by}")
        plt.ylabel(f"Presynaptic {group_by}")

    n_grp = np.maximum(len(common_dict["src_group_values"]), len(common_dict["tgt_group_values"]))
    font_size = max(13 - n_grp / 6, 1)  # Font scaling
    if len(common_dict["src_group_values"]) > 0:
        plt.yticks(
            range(len(common_dict["src_group_values"])),
            common_dict["src_group_values"],
            rotation=0,
            fontsize=font_size,
        )

    if len(common_dict["tgt_group_values"]) > 0:
        if max(len(str(grp)) for grp in common_dict["tgt_group_values"]) > 1:
            rot_x = 90
        else:
            rot_x = 0
        plt.xticks(
            range(len(common_dict["tgt_group_values"])),
            common_dict["tgt_group_values"],
            rotation=rot_x,
            fontsize=font_size,
        )

    cb = plt.colorbar()
    cb.set_label(res_dict["unit"])



# This file is part of connectome-manipulator.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) 2024 Blue Brain Project/EPFL

"""Module to facilitate easier job management"""

from contextlib import contextmanager
from pathlib import Path
from .. import log, utils, profiler


class JobTracker:
    """Class to track the execution of jobs"""

    def __init__(self, output_dir: Path, count: int):
        """Create a new tracker given an output directory and a job count"""
        self._jobs_done = 0
        self._jobs_total = count
        self._syn_count_in = 0
        self._syn_count_out = 0

        self.parquet_dir = output_dir / "parquet"
        self.parquet_done_file = self.parquet_dir / "parquet.DONE"

    def prepare_parquet_dir(self, resume: bool):
        """Setup and check the output parquet directory

        Will attempt to resume execution if `resume` is given, and will yield the job
        index and output file name for all jobs not yet process.
        """
        self.parquet_dir.mkdir(parents=True, exist_ok=True)

        # Potential list of .parquet file names [NOTE: Empty files won't actually exist!!]
        if self._jobs_total == 1:
            file_list = [self.parquet_dir / "edges.parquet"]
        else:
            ext_len = len(str(self._jobs_total))
            file_list = [
                self.parquet_dir / f"edges.{self._jobs_total}-{idx:0{ext_len}d}.parquet"
                for idx in range(self._jobs_total)
            ]

        # Check if parquet folder is clean
        existing_parquet_files = {p.stem for p in self.parquet_dir.glob("*.parquet")}
        done_list = set()
        if resume:
            # Resume from an existing run:
            # * Done file must be compatible with current run
            # * All existing .parquet files in the parquet folder must be from the list of expected (done) files (these files will be skipped over and merged later)
            if not self.parquet_done_file.exists():
                utils.write_json(data=[], filepath=self.parquet_done_file)
            else:
                # Load completed files from existing list [can be in arbitrary order!!]
                done_list = set(utils.load_json(self.parquet_done_file))
                unexpected = list(done_list - {p.stem for p in file_list})
                log.log_assert(
                    not unexpected,
                    f'Unable to resume! "{self.parquet_done_file}" contains unexpected entries: {unexpected}',
                )
            unexpected = list(existing_parquet_files - done_list)
            log.log_assert(
                not unexpected,
                f"Unable to resume! Parquet output directory contains unexpected .parquet files, please clean your output dir: {len(unexpected)} files",
            )
            # [NOTE: Empty files don't exist but may be marked as done!]
        else:  # Running from scratch: Parquet folder must not contain any .parquet files (so not to mix up existing and new files!!)
            log.log_assert(
                len(existing_parquet_files) == 0,
                'Parquet output directory contains .parquet files, please clean your output dir or use "do_resume" to resume from an existing run!',
            )
            utils.write_json(data=[], filepath=self.parquet_done_file)

        for i, file in enumerate(file_list):
            if file.stem in done_list:
                log.info(
                    f"Split {i + 1}/{self._jobs_total}: Parquet file already exists - "
                    f"SKIPPING (do_resume={resume})"
                )
            else:
                yield i, file

    def _mark_done(self, file):
        """Marks the given parquet file as "done" in the done file

        (i.e., adding the file name to the list of done files).
        """
        # Load list of completed files [can be in arbitrary order!!]
        done_list = utils.load_json(self.parquet_done_file)
        # Update list and write back
        done_list.append(file.stem)
        utils.write_json(data=done_list, filepath=self.parquet_done_file)

    @contextmanager
    def follow_jobs(self):
        """Start tracking jobs and provide a result hook to post-process them"""

        def _hook(result, info):
            self._syn_count_in += result[0]
            self._syn_count_out += result[1]
            self._mark_done(info["out_parquet_file"])
            profiler.ProfilerManager.merge(result[2])

            self._jobs_done += 1
            done_percent = self._jobs_done * 100 // self._jobs_total
            log.info(
                f"[{done_percent:3d}%] Finished {self._jobs_done} (out of {self._jobs_total}) splits"
            )

        yield _hook

        diff = self._syn_count_out - self._syn_count_in
        log.info("Done processing")
        log.info(
            "  Total input/output synapse counts: "
            f"{self._syn_count_in}/{self._syn_count_out} (Diff: {diff})"
        )
